/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.unitesk.requality.core.AttributeGenerator.ParameterAvailability;
import com.unitesk.requality.core.nodeiterators.AttributeValueIteratorType;

public class NodeMapper {
    private NodeMapper() {
    }

    public static UUID getUUId(String nodeString) throws JSONException {
        if (nodeString == null) {
            return null;
        }
        JSONObject obj = new JSONObject(nodeString);
        return UUID.fromString(obj.getString("uuid"));
    }

    public static Collection<Attribute> getAttributes(UUID uuid, String nodeString) throws JSONException {
        if (nodeString == null) {
            return null;
        }
        Collection<Attribute> result = new HashSet<Attribute>();
        JSONObject obj = new JSONObject(nodeString);
        try {
            JSONObject attrObj = obj.getJSONObject("attributes");
            loadNewFormat(uuid, result, attrObj);
        } catch (Exception e) {
            JSONArray attrObj = obj.getJSONArray("attributes");
            loadOldFormat(uuid, result, attrObj);
        }
        return result;
    }

    private static void loadOldFormat(UUID uuid, Collection<Attribute> result, JSONArray attr)
            throws JSONException {
        for (int i = 0; i < attr.length(); i++) {
            JSONObject a = attr.getJSONObject(i);
            AttributeType type = AttributeType.valueOf(a.getString("type"));
            String key = a.getString("key");
            Object value = null;
            if (type == AttributeType.LIST) {
                ArrayList<String> attrLst = new ArrayList<String>();
                JSONArray vals = a.getJSONArray("value");
                for (int j = 0; j < vals.length(); j++) {
                    attrLst.add(vals.getString(j));
                }
                value = attrLst;
            } else {
                // TODO: all attribute types support
                value = type.fromString(a.getString("value"));
            }
            result.add(new Attribute(uuid, type, key, value));
        }
    }

    private static void loadNewFormat(UUID uuid, Collection<Attribute> result, JSONObject attr)
            throws JSONException {
        for (String key : JSONObject.getNames(attr)) {
            JSONObject a = attr.getJSONObject(key);
        	UUID origin = uuid;
            AttributeType type = AttributeType.valueOf(a.getString("type"));
            Object value = null;
            AttributeType ltype = null;
            if (type == AttributeType.BOOL) {
                String v = a.getString("value");
                if ("true".equalsIgnoreCase(v) && !"false".equalsIgnoreCase(v)) {
                    value = Boolean.valueOf(v);
                } else {
                    value = v;
                }
            } else if (type == AttributeType.LIST) {
                value = listFromJSON(a);
            } else if (type == AttributeType.REFERENCE) {
            	value = a.getString("value");
            } else {
                // TODO: all attribute types support
                try {
                    value = type.fromString(a.getString("value"));
                } catch (Exception e) {
                    value = a.getString("value");
                }
            }
            
            try {
            	String or = a.getString("origin");
            	origin = UUID.fromString(or);
            } catch (Exception e) {
            }
            Attribute rattr = new Attribute(origin, type, key, value);
            if(a.has("generator")){
            	JSONObject obj = a.getJSONObject("generator");
            	Map<String,Object> setts = new HashMap<String, Object>();
            	JSONObject set = obj.getJSONObject("settings");
            	JSONArray names = set.names();
            	if(names!=null)
            		for(int i = 0;i<names.length();i++){
            			try{
            				JSONObject jo = set.getJSONObject(names.getString(i));
            				setts.put(names.getString(i), listFromJSON(jo));
            			}
            			catch(JSONException e){
            				setts.put(names.getString(i), set.get(names.getString(i)));	
            			}
            		}	
            	AttributeGenerator ag = new AttributeGenerator(rattr, UUID.fromString(obj.getString("uuid")),setts);
            	ag.setAvailability(ParameterAvailability.values()[obj.getInt("availability")]);
            	ag.setGeneratorType(AttributeValueIteratorType.fromId(obj.getString("generator_type")));
            	rattr = new Attribute(ag.getMasterUUId(),AttributeType.GENERATED,rattr.getKey(),rattr.getValue(),rattr.getRawValue(),ag);
            	//genObj.put("atribute_type", ag.getAttributeType().ordinal());
            	//genObj.put("atribute_list_type", ag.getAttributeListType().ordinal());
            }
            result.add(rattr);
        }
    }

    private static Object listFromJSON(JSONObject a) throws JSONException {
    	ArrayList<Object> attrLst = new ArrayList<Object>();
        JSONArray vals = a.getJSONArray("value");
        AttributeType ltype = AttributeType.STRING;
        if(a.has("elements_type"))
        	ltype = AttributeType.valueOf(a.getString("elements_type"));
        if(ltype != AttributeType.LIST)
        	for (int j = 0; j < vals.length(); j++) {
        		attrLst.add(ltype.fromString(vals.getString(j)));
        	}
        else
        {
        	for (int j = 0; j < vals.length(); j++) {
        		attrLst.add(listFromJSON(vals.getJSONObject(j))); 
        	}
        }
        return new Object[]{ltype, attrLst};
	}

	public final static JSONObject toJSON(TreeNode node) throws JSONException {
    	return toJSON(node.getUUId(), node.attributes.values());
    }

    @SuppressWarnings("unchecked")
    public final static JSONObject toJSON(UUID uuid, Collection<Attribute> attr)
            throws JSONException {
        JSONObject result = null;
        result = new JSONObject();
        result.put("uuid", uuid);
        JSONObject attrs = new JSONObject();
        for (Attribute a : attr) {
            JSONObject attrObj = new JSONObject();
            attrObj.put("isGenerated", a.isGenerated());
            attrObj.put("key", a.getKey());
            attrObj.put("type", a.getType());
            Object origin = a.getOrigin();
            if(origin == null)
            	origin = "";
            attrObj.put("origin", origin);
            if (a.getType() != AttributeType.LIST) {
            	attrObj.put("value", a.getRawValue());
            } else {
            	JSONArray lst = new JSONArray();
            	Object[]val = (Object[])a.getRawValue();
            	List<Object> al = (List<Object>) val[1];
            	AttributeType at = (AttributeType) val[0];
            	attrObj.put("elements_type", val[0]);
            	if(at!=AttributeType.LIST)
            		for (Object i : al) {
            			lst.put(i);
            		}
            	else
            		for (Object i : al) {
            			lst.put(listToJSON((Object[]) i));
            		}
            	attrObj.put("value", lst);
            }
            AttributeGenerator ag = a.getGenerator();
            if(ag!=null){
            	JSONObject genObj = new JSONObject();
            	genObj.put("uuid", ag.getUUID());
            	genObj.put("availability", ag.getAvailability().ordinal());
            	genObj.put("atribute_type", ag.getAttributeType().ordinal());
            	if(ag.getAttributeListType()!=null)
            		genObj.put("atribute_list_type", ag.getAttributeListType().ordinal());
            	genObj.put("generator_type", ag.getGeneratorType().getId());
            	JSONObject genAttrs = new JSONObject();
            	Map<String,Object>setts = ag.getSettings();
            	for(String s:setts.keySet())
            		if(setts.get(s) instanceof Object[])
            			genAttrs.put(s, listToJSON((Object[]) setts.get(s)));
            		else
            			genAttrs.put(s, setts.get(s));
            	genObj.put("settings", genAttrs);
            	attrObj.put("generator", genObj);
            }
            attrs.put(a.getKey(), attrObj);
        }
        result.put("attributes", attrs);
        return result;
    }

	private static JSONObject listToJSON(Object[] i) throws JSONException {
		JSONObject jo = new JSONObject();
        List<Object> al = (List<Object>) i[1];
        AttributeType at = (AttributeType) i[0];
        jo.put("elements_type", at);
        JSONArray lst = new JSONArray();
        if(at!=AttributeType.LIST)
        	for (Object o : al) {
        		lst.put(o);
        	}
        else
        	for (Object o : al) {
        		lst.put(listToJSON((Object[]) o));
        	}
        jo.put("value", lst);
		return jo;
	}

}
