/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.core;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;

import org.eclipse.core.commands.operations.IOperationHistory;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ISaveContext;
import org.eclipse.core.resources.ISaveParticipant;
import org.eclipse.core.resources.ISavedState;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.NodeDesc;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.NodeTypeDescriptor;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.eclipse.operations.TreeDBOperation;
import com.unitesk.requality.eclipse.storageformatversion.VersionUpdateInfo;
import com.unitesk.requality.eclipse.storageformatversion.StoredFormatVersioning;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.VirtualNode;

public class TreesTracker {
    public enum TreeEvent {
        TREE_OPEN, TREE_CLOSE, TREE_DELETE
    }

    private Map<IProject, TreeDB> treeByProject = new HashMap<IProject, TreeDB>();
    private Map<TreeDB, IProject> projectByTree = new HashMap<TreeDB, IProject>();
    private Map<IProject, ResourceStorage> projectToStorage = new HashMap<IProject, ResourceStorage>();

    private static TreesTracker tracker;
    private List<ITreeStateListener> listeners = new CopyOnWriteArrayList<ITreeStateListener>();

    public boolean addListener(ITreeStateListener l) {
        return listeners.add(l);
    }

    public boolean removeListener(ITreeStateListener l) {
        return listeners.remove(l);
    }

    private TreesTracker() {
    }

    public static TreesTracker getTracker() {
        if (tracker == null)
            tracker = new TreesTracker();
        return tracker;
    }

    public synchronized void notifyListeners(TreeDB db, TreeEvent type) {
        for (ITreeStateListener listener : listeners) {
            switch (type) {
            case TREE_CLOSE:
                listener.close(db);
                break;
            case TREE_OPEN:
                listener.open(db);
                break;
            case TREE_DELETE:
                listener.delete(projectByTree.get(db));
                break;
            }
        }
    }

    public IProject getProject(TreeDB db) {
    	IProject proj = projectByTree.get(db);
    	return proj;
    }

    public TreeDB getTree(IProject proj) {
        return treeByProject.get(proj);
    }

    private void showError(final String title, final String message) {
    	
    	new Thread(new Runnable() {
			@Override
			public void run() {
				while(!PlatformUI.isWorkbenchRunning())
					try {
						this.wait(200);
					} catch (Exception e) {}
				PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {
		    	    public void run() {
		    	    	Shell sh = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
						MessageDialog.openError(sh, title, message);
		    		}
		    	});
			}
		}).run();
	}
    private void showWarning(final String title, final String message) {
    	
    	new Thread(new Runnable() {
			@Override
			public void run() {
				while(!PlatformUI.isWorkbenchRunning())
					try {
						this.wait(200);
					} catch (Exception e) {}
				PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {
		    	    public void run() {
		    	    	Shell sh = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
						MessageDialog.openWarning(sh, title, message);
		    		}
		    	});
			}
		}).run();
	}
    private void closeProject(final IProject project){
    	Job job = new Job("Close project Job") {
    		  @Override
    		  protected IStatus run(IProgressMonitor monitor) {
    			  while(!PlatformUI.isWorkbenchRunning())
  					try {
  						this.wait(200);
  					} catch (Exception e) {}
  					PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {
  			    	    public void run() {
  			    	    	try {
  			    	    		project.close(null);
  			    	    	} catch (CoreException e) {
  			    	    		e.printStackTrace();
  			    	    	}
  			    	    }
  					});
    		    return Status.OK_STATUS;
    		  }
    		};
    		job.setPriority(Job.DECORATE);
    		job.schedule(200);
    }
    private List<IProject> locked = new ArrayList<IProject>();
    public boolean trackProject(final IProject project) {
        try {
        	if(!project.isOpen() || !project.isNatureEnabled(ReqCatalogue.ID)){
        		return false;
        	}else{
                TreeDB db = getTree(project);
                if (db == null) {
                	if(!locked.contains(project)){
                		List<VersionUpdateInfo> vilist = StoredFormatVersioning.collectUpdates(project);
	                	StringBuilder updates = new StringBuilder("");
	                	StringBuilder plugs = new StringBuilder("");
	                	String problems = "",p2 = null;
	                	int pl = 0;
	    	        	for(VersionUpdateInfo vi : vilist){
	    	        		if(vi.getConverter() == null){
	    	        			pl++;
	    	        			if(vi.getStatus().equals("plugin_not_found")){
	    	        				plugs.append("\n");
	    	        				plugs.append(vi.getGroupId());
	    	        			}else
	    	        				p2 = vi.getNewVersion();
    	        				problems = vi.getStatus();
	    	        		}else{
	    	        			updates.append("\n");
	    	        			updates.append(vi.toString());
	    	        		}
	    	        			
	    	        	}
	    	        	if (pl > 0)
	    	        	{
	    	        		locked.add(project);
	    	        		if(problems.equals("version_not_found"))
	    	        		{
	    	        			showError("Project Loading","Format of the project '"+project.getName()+"' requires newer version of Requality ("+p2+").");
	    	        		}else
	    	        		if(problems.equals("plugin_not_found"))
		    	        	{
	    	        			pl--;
	    	        			plugs.append("\n");
	    	        			showWarning("Project Loading: '"+project.getName()+"'","Current set of plugins cannot handle the following elements: "+
	    	        					plugs.toString()+
	    	        					"Install additional plugins.");
	    	        		}
	    	        	}
	    	        	if(pl>0){
	    	        		closeProject(project);
	    	        		locked.remove(project);
	    	        		return false;	
	    	        	}
	    	        	else
	    	        	{
	    	        		if(updates.length() > 0)
	    	        		{
	    	        			locked.add(project);
	    	        			showQuestion("Project Loading"
	    	        					, "Format of the project '"+project.getName()+"' is out of date, and installed Requality version can't process it. Do you want to update the project to proper format?"
	    	        					, new Runnable() {
	    									@Override
	    									public void run() {
	    										List<VersionUpdateInfo> res = StoredFormatVersioning.updateProject(project);
	    										int rs = res.size();
	    										if(res.size() > 0)
	    										{
	    											StringBuilder sb = new StringBuilder();
	    											String problems = "", pr2 = "";
	    											for(VersionUpdateInfo vui:res)
	    												if(vui.getConverter() == null){
	    													problems = vui.getStatus();
	    													if(problems.equals("plugin_not_found")){
	    														sb.append("\n");
	    														sb.append(vui.getGroupId());
	    														rs--;
	    													}else
	    														pr2 = vui.getNewVersion();
	    												}
	    											if(problems.equals("version_not_found"))
	    							    	        {
	    							    	        	showError("Project Loading","Format of the project '"+project.getName()+"' requires newer version of Requality ("+pr2+").");
	    							    	        }else if(problems.equals("plugin_not_found"))
	    							    	        {
	    							    	        	showWarning("Project Loading: '"+project.getName()+"'","Current set of plugins cannot handle the following elements: "+
	    							    	        			sb.toString()+
	    							    	        			"Install additional plugins.");
	    							    	        }else{
	    							    	        	showError("Project Loading","Conversion of '"+project.getName()+"' to the new format failed.");
	    							    	        }
	    										}
	    										if(rs == 0){
	    											trackProject(project);
	    										}
	    										locked.remove(project);
	    									}
	    								},
	    								new Runnable() {
	    									public void run() {
	    										try {
	    											project.close(null);
	    										} catch (CoreException e) {
	    											e.printStackTrace();
	    										}
	    										locked.remove(project);
	    									}
	    								});
	    	        			//errors still may appear in project
	    	        			return false;
	    	        		}
	    	        	}
                	}
                	if(!project.isOpen())
                		return false;
                    db = loadTreeDB(project);
                    Map<String, NodeTypeDescriptor> nts = RequalityPlugin
                            .getAdditionalTypes();
                    bindBaseToProject(project, db);
                    for (String name : RequalityPlugin.getAdditionalTypes()
                            .keySet()) 
                        db.addNodeType(name, nts.get(name).getClazz());
                    db.addListener(new DefaultTreeChangeListener() {
                        @Override
                        public void deleted(NodeDesc node) {
                        }
                        @Override
                        public void created(TreeNode node){
                        	if(node.getGroup() != TreeNode.NODE_GROUP){
                            	IProject proj = TreesTracker.getTracker().getProject(node.getTreeDB());
                            	List<String> vers = StoredFormatVersioning.getFormatsVersions(proj);
                            	String v = StoredFormatVersioning.getFullVersionByGroupId(node.getGroup());
                            	if(!vers.contains(v))
                            	{
                            		vers.add(v);
                            		StoredFormatVersioning.setFormatsVersions(proj, vers);
                            	}
                            }
                        }
                    });
                    notifyListeners(db, TreeEvent.TREE_OPEN);

                    ResourcesPlugin.getWorkspace().addResourceChangeListener(
                            ((ResourceStorage) db.getStorage()).getResourceChangeListener());
                    WorkspaceJob runBuilder = new WorkspaceJob("Check Requality Project: "
                            + project.getName()) {
                        @Override
                        public IStatus runInWorkspace(IProgressMonitor monitor)
                                throws CoreException {
                            try {
                                project.build(IncrementalProjectBuilder.FULL_BUILD, monitor);
                            } catch (CoreException e) {
                                e.printStackTrace();
                            }
                            return Status.OK_STATUS;
                        }
                    };
                    runBuilder.setRule(ResourcesPlugin.getWorkspace().getRuleFactory().buildRule());
                    runBuilder.schedule();
                }
                if(db!=null)
                	Location.clearCache(db);
            }
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    private void showQuestion(final String title, final String message, final Runnable runnable,
			final Runnable runnable2) {
    	new Thread(new Runnable() {
    		@Override
    		public void run() {
    			while(!PlatformUI.isWorkbenchRunning())
    				try {
    					this.wait(200);
    				} catch (Exception e) {}
    			PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {
    	    	    public void run() {
    	    	    	Shell sh = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
    	    	    	if(MessageDialog.openConfirm(sh, title, message)){
    	    	    		if(runnable != null)
    	    	    			runnable.run();
    	    	    	}else{
    	    	    		if(runnable2 != null)
    	    	    			runnable2.run();
    	    	    	}
    	    	    }
    			});
    		}
    	}).run();
	}

	private TreeDB loadTreeDB(final IProject project) {
        if(!projectToStorage.containsKey(project)) {
            new ResourceStorage(project);
        }
        ResourceStorage storage = projectToStorage.get(project);
        final TreeDB db = TreeDB.load(storage);
        db.addTransactionListener(new ITransactionListener() {
            public IUndoContext getContext() {
                return PlatformUI.getWorkbench()
                        .getOperationSupport().getUndoContext();
            }

            @Override
            public void _dispose() {
                //IOperationHistory history = PlatformUI.getWorkbench()
                //        .getOperationSupport().getOperationHistory();
                //history.dispose(getContext(), true, true, true);
            }

            @Override
            public void execute(TransactionStorage tStorage) {
                TreeDBOperation operation = new TreeDBOperation(tStorage);

                operation.addContext(getContext());
                IOperationHistory history = PlatformUI.getWorkbench()
                        .getOperationSupport().getOperationHistory();
                history.add(operation);
            }

            @Override
            public void redo(TransactionStorage tStorage) {
            }

            @Override
            public void undo(TransactionStorage tStorage) {
                // TODO Auto-generated method stub

            }
        });

        ISaveParticipant saveParticipant = new ISaveParticipant() {
			@Override
			public void saving(ISaveContext context) throws CoreException {
				VirtualNode.getVirtualStorage(db).saveChanges();
			}
			
			@Override
			public void rollback(ISaveContext context) {
				// TODO Auto-generated method stub
				
			}
			
			@Override
			public void prepareToSave(ISaveContext context) throws CoreException {
			}
			
			@Override
			public void doneSaving(ISaveContext context) {
				// TODO Auto-generated method stub
				
			}
		};
        try {
			ISavedState lastState =
			   ResourcesPlugin.getWorkspace().addSaveParticipant(RequalityPlugin.ID, saveParticipant);
		} catch (CoreException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        return db;
    }

    public boolean stopTracking(IProject project, boolean deleted) {
        if (!isTracking(project)) {
            return false;
        }
        TreeDB db = getTree(project);
        projectToStorage.remove(project);
        if(db != null){
        	ITreeStorage ts = db.getStorage();
        	//TODO error?
        	if(ts instanceof TransactionStorage)
        		ts = ((TransactionStorage)ts).getStorage();
            ResourceStorage storage = (ResourceStorage) ts;
            ResourcesPlugin.getWorkspace().removeResourceChangeListener(
                storage.getResourceChangeListener());

            if (!deleted) {
                notifyListeners(db, TreeEvent.TREE_CLOSE);
                unbindBaseFromProject(project, db);
                treeByProject.put(project, null);
            } else {
                notifyListeners(db, TreeEvent.TREE_DELETE);
                unbindBaseFromProject(project, db);
            }
            return true;
        }else{
            return false;
        }
    }

    private void unbindBaseFromProject(IProject project, TreeDB db) {
        projectByTree.remove(db);
        treeByProject.remove(project);
    }

    private void bindBaseToProject(IProject project, TreeDB db) {
        projectByTree.put(db, project);
        treeByProject.put(project, db);
    }

    public boolean isTracking(IProject project) {
        boolean result = treeByProject.containsKey(project);
        return result;
    }

    public TreeDB[] getTrees() {
        return projectByTree.keySet().toArray(new TreeDB[0]);
    }

    public IProject[] getProjects() {
        return projectByTree.values().toArray(new IProject[0]);
    }

    private boolean isTrackable(IProject project) {
        try {
            return project.isNatureEnabled(ReqCatalogue.ID);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    public void setProjects(IProject[] projects) {
        HashSet<IProject> current = new HashSet<IProject>(
                treeByProject.keySet());

        for (IProject project : projects) {

            try {
                if (isTrackable(project) && !isTracking(project)) {
                    trackProject(project);
                } else if (!isTrackable(project) && isTracking(project)) {
                    stopTracking(project, project.exists());
                }
                current.remove(project);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        for (IProject iProject : current) {
            stopTracking(iProject, iProject.exists());
        }
    }

    public static TreeNode getNodeByResource(IFile file) {
        TreeDB db = getTracker().getTree(file.getProject());
        if(db == null)
        	return null;
        ResourceStorage rs = getResourceStorage(db);
        UUID uuid = rs.getUUId(file);
        if (uuid == null) {
            return null;
        }
        try{
        	return db.getNode(uuid);
        }catch(RuntimeException re){
        	//node not found
        	return null;
        }
    }

    public static IResource getResourceByNode(TreeNode node) {
    //TODO check it
    	TreeDB db = node.getTreeDB();
    	UUID uid = node.getUUId();
    	if(node.isVirtual())
    		return null;
    	ResourceStorage rs = getResourceStorage(db);
    	if (rs != null)
    		return rs.getResource(uid);
    	
    	return null;
    }

    public static ResourceStorage getResourceStorage(TreeDB db) {
        ResourceStorage rs = null;
        if (db.getStorage() instanceof TransactionStorage) {
        	if(((TransactionStorage) db.getStorage())
                    .getStorage() instanceof ResourceStorage)
        		rs = (ResourceStorage) ((TransactionStorage) db.getStorage())
                    	.getStorage();
        } else if(db.getStorage() instanceof ResourceStorage){
            rs = (ResourceStorage) db.getStorage();
        }
        return rs;
    }

    public void assignTreeDB(IProject project, ResourceStorage resourceStorage) {
        projectToStorage.put(project, resourceStorage);
    }
}
