/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.NodeDesc;
import com.unitesk.requality.core.TreeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.core.TreesTracker;

public class RequalityContentLinksProvider implements ITreeContentProvider {
    private TreeViewer viewer;
    private ArrayList<TreeDB> dbs = new ArrayList<TreeDB>();

    public RequalityContentLinksProvider() {
    }

    private ITreeChangeListener treeChangeListener = new TreeChangeListener() {
        private List<TreeNode> dirtyNodes = new ArrayList<TreeNode>();
        
        private void guiRefresh(final TreeNode node) {
            if (dirtyNodes.contains(node) || viewer.getControl().isDisposed()) {
                return;
            }
            dirtyNodes.add(node);
            viewer.getControl().getDisplay().asyncExec(new Runnable() {
                @Override
                public void run() {
                    dirtyNodes.remove(node);
                    if (viewer.getControl().isDisposed()) {
                        return;
                    }
                    Object o = viewer.getInput();
                    //System.out.println(node.toString());
                    if(o == null || !(o instanceof TreeNode) ||  node == null ||((TreeNode)o).getUUId().equals(node.getUUId()))
                    	viewer.refresh(true);
                    if(node!=null && !node.isDisposed()&&!viewer.isBusy())
                    	viewer.refresh(node,true);
                }
            });
        }

        @Override
        public void moved(TreeNode node, TreeNode oldParent) {
            guiRefresh(node);
            guiRefresh(oldParent);
        }

        @Override
        public void deleted(NodeDesc node) {
            guiRefresh(null);
        }

        @Override
        public void changed(TreeNode node) {
            guiRefresh(node);
        }

        @Override
        public void created(TreeNode node) {
            guiRefresh(node);
        }
    };

    private ITreeStateListener stateListener = new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
            dbs.add(db);
            db.addGUIListener(treeChangeListener);
        }

        @Override
        public void delete(IProject proj) {
        	dbs.remove(TreesTracker.getTracker().getTree(proj));
        }

        @Override
        public void close(TreeDB db) {
            dbs.remove(db);
        }
    };

    @Override
    public void dispose() {
        RequalityPlugin.getTreesTracker().removeListener(stateListener);
    }

    @Override
    public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
        RequalityPlugin.getTreesTracker().removeListener(stateListener);
        if (!(viewer instanceof TreeViewer)) {
            throw new IllegalArgumentException();
        }
        this.viewer = (TreeViewer) viewer;
        for (TreeDB db : dbs) {
            db.removeGUIListener(treeChangeListener);
        }
        dbs.clear();

        RequalityPlugin.getTreesTracker().addListener(stateListener);
        if (newInput == null) {
            return;
        }
        if (newInput instanceof TreeNode) {
            dbs.add(((TreeNode) newInput).getTreeDB());
        } else if (newInput instanceof IWorkspaceRoot) {
            dbs.addAll(Arrays.asList(RequalityPlugin.getTreesTracker()
                    .getTrees()));
        } else if (newInput instanceof Object[]) {
            dbs.add(((TreeNode) ((Object[]) newInput)[0]).getTreeDB());
        } else if (newInput instanceof TreeDB) {
            dbs.add((TreeDB) newInput);
        } else {
            throw new IllegalArgumentException();
        }
        for (TreeDB db : dbs) {
            db.addGUIListener(treeChangeListener);
        }
        viewer.getControl().getDisplay().asyncExec(new Runnable() {
            @Override
            public void run() {
                //TODO if control is disposing, old code will throw exception
                if((RequalityContentLinksProvider.this.viewer.getControl() != null) && 
                        !RequalityContentLinksProvider.this.viewer.getControl().isDisposed())
                    RequalityContentLinksProvider.this.viewer.refresh();
            }
        });
    }
    public static class TestNodeLinks{
    	private String name;
    	private Object[] nodes;
    	private TreeNode target;
    	private boolean isup;
    	public TestNodeLinks(boolean up, TreeNode node, String name, Object[] linked){
    		nodes = linked;
    		this.name = name;
    		target = node;
    		isup = up;
    	}
    	public boolean isUp(){
    		return isup;
    	}
    	public Object[]getChildren(){
    		return nodes;
    	}
    	@Override
    	public String toString() {
    		return name;
    	}
		public TreeNode getTarget() {
			return target;
		}
    }

    @Override
    public Object[] getElements(Object inputElement) {
        return getChildren(inputElement);
    }
    
    private TreeNode getTreeNode(Object parentElement){
    	TreeNode parent = null;
    	if (parentElement instanceof IProject) {
            IProject project = (IProject) parentElement;
            if (!project.isOpen()) {
                return null;
            }
            if (TreesTracker.getTracker().getTree(project) != null
                    && TreesTracker.getTracker().getTree(project).getRootNode() != null // Root
                                                                                        // node
                                                                                        // can
                                                                                        // be
                                                                                        // not
                                                                                        // created
                                                                                        // yet
            ) {
                parent = TreesTracker.getTracker().getTree(project)
                        .getRootNode();
            }
        }
        if (parentElement instanceof TreeNode) {
            parent = (TreeNode) parentElement;
        }
        if (parentElement instanceof TreeDB) {
            parent = ((TreeDB) parentElement).getRootNode();
        }
        return parent;
    }

    private void addListReferences(TreeNode parent,String key, Map<String,Object> links, Object[] list){
    	List l = (List)list[1];
    	if(list[0] == AttributeType.LIST){
    		for(int i=0;i<l.size();i++)
    			addListReferences(parent,key+"["+i+"]", links,(Object[])l.get(i));
    	}else if(list[0] == AttributeType.REFERENCE){
    		for(int i=0;i<l.size();i++){
    			TreeNode node = parent.resolveLink((String)l.get(i), null, key+"["+i+"]");
    			if(node!=null)
    				links.put(key+"["+i+"]",node);
    		}
    	}
    }
    @Override
    public Object[] getChildren(Object parentElement) {
    	if (parentElement instanceof IProject) {
            IProject project = (IProject) parentElement;
            if (!project.isOpen()) {
                return new Object[0];
            }
    	}
    	if(parentElement instanceof TreeNode){
    		TreeNode parent = getTreeNode(parentElement);
    		Map<String,Object> links = new HashMap<String,Object>(),rlinks = new HashMap<String, Object>();
    		for(Attribute a:parent.getAttributes().values())
    			if(a!=null){
    				if(a.getType().equals(AttributeType.REFERENCE))
    					links.put(a.getKey(),parent.getLinkAttributeValue(a.getKey()));
    				if(a.getType().equals(AttributeType.LIST))
    					addListReferences(parent,a.getKey(),links,(Object[])a.getRawValue());
    			}
    		Map<TreeNode,Set<String>>revlinks = parent.getRevertLinks();
    		for(TreeNode nd:revlinks.keySet())
    			for(String s:revlinks.get(nd)){
    				if(!rlinks.containsKey(s))
    					rlinks.put(s, new ArrayList());
    				((ArrayList) rlinks.get(s)).add(nd);
    			}
    		List<Object>tnl = new ArrayList<Object>();
    		//TODO 
    		for(String s:links.keySet())
   				tnl.add(new TestNodeLinks(false, parent, s, new Object[]{links.get(s)}));
    		for(String s:rlinks.keySet())
   				tnl.add(new TestNodeLinks(true, parent, s, ((List)rlinks.get(s)).toArray()));
        	return tnl.toArray();
    	}else if (parentElement instanceof TestNodeLinks)
    		return ((TestNodeLinks)parentElement).getChildren();
    	return new Object[0];
    }

    @Override
    public Object getParent(Object element) {
    	if(element instanceof TestNodeLinks){
    		return ((TestNodeLinks)element).getTarget();
    	}
        if (element instanceof TreeNode) {
            TreeNode node = (TreeNode) element;
            if (node.isDisposed()) {
                return null;
            }
            if (node.getTreeDB().getRootNode().equals(node.getParent())) {
                return TreesTracker.getTracker().getProject(node.getTreeDB());
            }
            return node.getParent();
        }
        return null;
    }

    @Override
    public boolean hasChildren(Object element) {
    	return getChildren(element).length>0;
    }
}
