/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.views.documents;

import org.w3c.dom.*;

import freemarker.core.LocalContext;

import java.util.*;
import java.util.regex.Pattern;


public class DocStringRepresentation {

    protected Document doc;
    protected StringBuilder strRep = new StringBuilder();
    protected List<SpanLocation> spans = new ArrayList<SpanLocation>();
    protected List<NodeLocation> tagsList = new ArrayList<NodeLocation>();
    protected Map<Node, LocationPosition> tagsMap = new HashMap<Node, LocationPosition>();
    protected List<Integer> tagStartToTagNumber = new ArrayList<Integer>();
    
    protected Set<String> addedSpanIds = new HashSet<String>();
    int transferredSpansCounter = 0;
    int foundButNotTransferredCounter = 0;
    int foundInFewNodesCounter = 0;
    int transferredFromNextDOMTags = 0;
    
    public Map<Node, Node> debugRemovedNodes = new HashMap<Node, Node>();
	private boolean verbose = false;
    
    public DocStringRepresentation(final Document doc) {
        this.doc = doc;
        buildString(doc, true);
        if (verbose ) {
        	System.out.println("spans size = " + spans.size());
        }
        Collections.sort(tagsList);
        for (int i = 0; i < tagsList.size(); ++i) {
        	tagStartToTagNumber.add(tagsList.get(i).location.start);
        }
        
    }
    
    public void buildString(final Node temp, boolean removeNewLines) {
        int start = strRep.length();
		boolean appendNewLine = false;
        
        if (temp.getNodeType() == Node.TEXT_NODE) {
        	String text = temp.getNodeValue();
        	append(text, removeNewLines);
        } else {
        	Pattern BLOCK_ELEMENT_RE = Pattern.compile("p|h[0-9]|br|div|tr|td", Pattern.CASE_INSENSITIVE);
        	String nodeName = temp.getNodeName();
        	if (temp.getNodeType() == Node.ELEMENT_NODE && BLOCK_ELEMENT_RE.matcher(nodeName).matches()) {
        		removeNewLines = true;
        		
        		appendNewLine = true;
        	}
        			
            NodeList nl = temp.getChildNodes();

            for (int i = 0; i < nl.getLength(); i++) {
                buildString(nl.item(i), removeNewLines);
            }

            if (temp.getNodeName().compareTo("span") == 0) {
				Node clazz = temp.getAttributes().getNamedItem("class");
				if (clazz != null) {
	                String spanId = clazz.getNodeValue();
	                if (spanId.startsWith("requality_text ")) {
	                	spans.add(new SpanLocation(new LocationPosition(start, strRep.length()), spanId, strRep.substring(start)));
	                }
				} 
            }
            if (appendNewLine) {
            	appendNewLine();
            } 
        }
        
        LocationPosition newLoc = new LocationPosition(start, strRep.length());

    	NodeLocation nodeLoc = new NodeLocation(newLoc, temp); 
        tagsList.add(nodeLoc);
        
        
        tagsMap.put(temp, newLoc);
        
        
    }

	private void append(String text, boolean removeNewLines) {
		if (removeNewLines) {
//			text = text.replaceAll("\r?\n", " ").replaceAll("\t|\\s{2,}", " ")/*.trim()*/;
			text = text.replaceAll("\r?\n", " ").replaceAll("\t", " ").replaceAll("\\s{2,}", " "); //.replaceAll("\\s[,\\.]", "!");
			text = text.replaceAll("\\s[\\.]", ".").replaceAll("\\s[,]", ",");
//			text = text.replaceAll("\r?\n", " ").replaceAll("\t", " ").replaceAll("\\s{2,}", " ").replaceAll("\\s,", ",");
		}
		if (strRep.length() > 0 && (lastChar() == ' '/* || lastChar() == '\n'*/) ) {
			text = text.replaceFirst("^\\s+", "");
		}
		
		if (strRep.length() > 0 && (lastChar() == ' '/* || lastChar() == '\n'*/) ) {
//			text = text.replaceFirst("^[,\\.]", "");
			if (text.startsWith(",") || text.startsWith(".") || text.startsWith(";") ) {
				strRep.deleteCharAt(strRep.length() - 1);
			}
		}
		strRep.append(text);
	}
	private void appendNewLine() {
//		if (strRep.length() > 0 && (lastChar() == ' '/* || lastChar() == '\n'*/) ) {
//			strRep.setLength(strRep.length()-1);
//		}
		strRep.append("\r\n");
	}
	
	private int getTagNumber(int positionStart) {
		
		int low = 0;
		int last = tagStartToTagNumber.size() - 1;
		int high = last;
		
		while (low <= high) {
			int mid = low + (high - low) / 2;
			int midVal = tagStartToTagNumber.get(mid);
			
			int next = (mid < last) ? tagStartToTagNumber.get(mid + 1) : tagStartToTagNumber.get(last);
			if ( (midVal <= positionStart) && (positionStart <= next) ) {
				low = mid;
				break;
			}
			
			if (positionStart < midVal) {
				high = mid - 1;
			} else if (positionStart > midVal) {
				low = mid + 1;
			}
		}
		
		// moving to first DOM node with the same start
		if (low > last) {
			low = last;
		}
		while ((low > 0 ) && (tagStartToTagNumber.get(low).equals(tagStartToTagNumber.get(low - 1))) ) {
			--low;
		}
		
		return low ;
	}
	
	
	private char lastChar() {
		return strRep.charAt(strRep.length()-1);
	}
    
    public String getCleanString() {
        return strRep.toString();
    }
    
    public List<SpanLocation> getSpans() {
        return spans;
    }
    
    public List<NodeLocation> getTags() {
        return tagsList;
    }
    
    
    public boolean DocModify(final LocationToTransfer locationPositionTextInNewFile, final String spanId) {
    	Node node = null;
        String spanText = strRep.toString().substring(locationPositionTextInNewFile.getLocation().getStart(),  
        											  locationPositionTextInNewFile.getLocation().getEnd());
               
        if (spanText.trim().length() == 0) {
        	return false;
        }
        
        int domTagNumber = getTagNumber(locationPositionTextInNewFile.getLocation().start);
        
        node = tagsList.get(domTagNumber).getNode();
        if (node == null) {
            return false;
        }
      
        boolean result = false;
        
        if (node.getNodeValue() != null && 
           (node.getNodeValue().contains(spanText) || node.getNodeValue().replaceAll("\\s{2,}", " ").contains(spanText)) ) {
	        if (!node.hasChildNodes() && node.getNodeName().compareTo("#text") == 0) {
	        	result = updateForSpanNodes(locationPositionTextInNewFile, spanId, node, spanText);
	        	return result;
	        } 
	        return false;
	        
        } else {
        	// TODO: append few tags from tagList and update few nodes 
        	
        	++foundButNotTransferredCounter;
        	int offset = 1;
        	if ( (domTagNumber + offset) >= tagsList.size() ) {
        		return false;
        	}
        	Node nextNode = tagsList.get(domTagNumber + offset).getNode();
        	
        	String strToTransferSpan = "" ;
        	
        	int counter = 0;
        	while (!strToTransferSpan.replaceAll("[\\r\\n\\s]+", " ").contains(spanText)) {
        		if (verbose) {
            		System.out.println("strToTransferSpan = " + strToTransferSpan.replaceAll("[\\r\\n\\s]+", " ") 
            				+ " === span = " + spanText);
        		}

        		++counter;
        		if ( (domTagNumber + offset) >= tagsList.size() ) {
            		return false;
            	}
        		nextNode = tagsList.get(domTagNumber + offset).getNode();
        		if (nextNode.getNodeValue() != null) {
        			if (!nextNode.getNodeValue().equals("\n")) {
        				strToTransferSpan += nextNode.getNodeValue();
        			}
        		}
        		++offset;
        		
        		if (counter > 25) {
        			if (verbose) {
        				System.out.println("Text has not been found in 25 appended nodes");
        			}
        			return false;
        		}
        	}
        	foundInFewNodesCounter++;
        	
        	// understanding how many nodes we should change
        	// currently working if span is only in one node
        	int numberOfNodesToUpdate = 0;
        	String str = "";
        	for (int i = offset - 1; i >= 0; --i) {
        		Node last = tagsList.get(domTagNumber + i).getNode();
        		str = last.getNodeValue() + str;
        		++numberOfNodesToUpdate;
        		if (str.contains(spanText) || str.replaceAll("[\\r\\n\\s]+", " ").contains(spanText)) {
        			break;
        		}
        	}
        	
        	if (numberOfNodesToUpdate == 1) {
        		nextNode = tagsList.get(domTagNumber + offset - 1).getNode();
        		result = updateForSpanNodes(locationPositionTextInNewFile, spanId, nextNode, spanText);
        		if (result) {
        			++transferredFromNextDOMTags;
        		}
        		
        	} else {
        		System.out.println("TODO: update few DOM nodes");
        		// TODO update few nodes in DOM
        	}
        	
        	return result;
        }
    }

	private boolean updateForSpanNodes(final LocationToTransfer locationPositionInNewFile,
			final String spanId, Node node, String spanText) {
		
		Node currentNode = node;
		
		while (debugRemovedNodes.keySet().contains(currentNode)){
			currentNode = debugRemovedNodes.get(currentNode);
		}

		if (verbose) {
			System.out.println("currentNode = " + currentNode);
			System.out.println("spanText = <" + spanText + ">");
		}
		
		if (currentNode == null) {
			if (verbose) {
				System.out.println("in updateForSpanNodes for span <" + spanText + ">, in if block with currentNode is null");
			}
			// if here - than the text of next span is contains in one of the previous, so currently we find wrong location
			return false;
		}
		
		LocationPosition locationOfNodeInDoc = tagsMap.get(currentNode);
		
		if (locationOfNodeInDoc == null) {
			if (verbose) {
				System.out.println("returning false as tagsMap doesn't contain currentNode");
			}
			return false;
		}
		
		if ( ! ((locationPositionInNewFile.getLocation().getStart() >= locationOfNodeInDoc.getStart()) ||
			     (locationOfNodeInDoc.getEnd() >= locationPositionInNewFile.getLocation().getEnd()) )) {
				return false;
		}
		
		LocationPosition occurencePosition = getOccurence(locationPositionInNewFile, locationOfNodeInDoc, currentNode, spanText);
		if (occurencePosition == null) {
			return false;
		}
		
		Node left = null;
		try {
			left = doc.createTextNode(currentNode.getTextContent().substring(0, occurencePosition.getStart()));
		} catch (StringIndexOutOfBoundsException e ) {
			left = null;
		}
		
		Node right = null;
		try {
			right = doc.createTextNode(currentNode.getTextContent().substring(occurencePosition.getEnd()));
		} catch (StringIndexOutOfBoundsException e) {
			right = null;
		}
		
		Element newNode = doc.createElement("span");
		newNode.setAttribute("class", spanId);
		Node newTextNode = doc.createTextNode(spanText);
		newNode.appendChild(newTextNode);
		
		Node parent = currentNode.getParentNode();
		
		if (left != null) {
			parent.insertBefore(left, currentNode);
		}
		
		parent.insertBefore(newNode, currentNode);
		++transferredSpansCounter;
		boolean isFirstSpanOccurence = addedSpanIds.add(spanId);
		if (isFirstSpanOccurence) {
			// adding tag <a>
			Element tagA = doc.createElement("a");
			
			tagA.setAttribute("class", "requality_id");
			tagA.setAttribute("id", spanId.substring(15));
			tagA.setAttribute("name", spanId.substring(18));
			
		    Node emptyTextNode = doc.createTextNode("");
		    tagA.appendChild(emptyTextNode);
		    newNode.insertBefore(tagA, newTextNode);
		}
		
		if (right != null)
		{
			parent.insertBefore(right, currentNode);
		}
				
		debugRemovedNodes.put(currentNode, right);
			
		int newEnd = locationOfNodeInDoc.getEnd();
		
		LocationPosition modifiedLoc = new LocationPosition(newEnd - right.getTextContent().length(), newEnd);

		// modify location in tagsMap
		tagsMap.put(right, modifiedLoc);
		if (verbose) {
			System.out.println("Changing tagsMap in:");
			System.out.println("        [" + right + "] -> [" + modifiedLoc + "]");
		}
		
		return true;
	}
    
    private LocationPosition getOccurence(LocationToTransfer locationPositionInNewFile, LocationPosition locationOfNodeInDoc, Node currentNode, String spanText) {
    	int startOccurenceIndex = locationPositionInNewFile.getLocation().getStart() - locationOfNodeInDoc.getStart();
		int endOccurenceIndex =	locationPositionInNewFile.getLocation().getEnd() - locationOfNodeInDoc.getStart();
        
        if (startOccurenceIndex < 0 || endOccurenceIndex < 0) {
        	if (verbose) {
        		System.out.println("Error while finding indexes of spanText in DOM node text, ignore this span");
        	}
        	return null;
        }
		
        try {
        	String spanString = currentNode.getTextContent().substring(startOccurenceIndex, endOccurenceIndex);
        	if (verbose) {
        		System.out.println("spanString = <" + spanString + ">");
        		System.out.println("spanText = <" + spanText + ">");
        	}
        	
        	if (! (spanString.replaceAll("[\\r\\n\\s]+", " ")).equals(spanText)) {
        		if (verbose) {
        			System.out.println("spanText and SpanString are not equals");
        		}
        		
        		String nodeTextContentWithoutMultipleSpaces = currentNode.getTextContent().replaceAll("\\s{2,}", " ");
        		
        		if (nodeTextContentWithoutMultipleSpaces.contains(spanText)) {
        			// we should modify start end end indexes for update in DOM of new document
        			if (verbose) {
        				System.out.println("Trying to transfer this requirement");
        			}
        			
        			int fromIndex = nodeTextContentWithoutMultipleSpaces.indexOf(spanText);
        			if (fromIndex == -1) {
        				return null;
        			}
        			// check for another occurrence  - if found - it is ambiguous, ignore it
        			if (fromIndex < nodeTextContentWithoutMultipleSpaces.length() && 
        					nodeTextContentWithoutMultipleSpaces.indexOf(spanText, fromIndex + 1) != -1) {
        				if (verbose) {
        					System.out.println("Ambiguous occurence");
        				}
        				return null;
        			}
        			
        			String firstSpanTextWord = spanText.split("\\s")[0];
        			startOccurenceIndex = currentNode.getTextContent().indexOf(firstSpanTextWord, fromIndex);
        			fromIndex += spanText.length();
        			
        			String[] wordsArray = spanText.split("\\s");
        			int len = wordsArray[wordsArray.length - 1].length();
        			String lastSpanTextOccur =  spanText.substring(spanText.length() - len);
        			endOccurenceIndex = currentNode.getTextContent().indexOf(lastSpanTextOccur, fromIndex - len) + lastSpanTextOccur.length();
        			
    				String newSpanString = currentNode.getTextContent().substring(startOccurenceIndex, endOccurenceIndex);
    				if (verbose) {
    					System.out.println("newSpanString = <" + newSpanString + ">");
    				}
    				int spanTextLen = spanText.length();
    				int newSpanStringLen = newSpanString.length();
    				
    				if (! newSpanString.replaceAll("[\\r\\n\\s]+", " ").equals(spanText) ||
        					! newSpanString.substring(0,1).equals(spanText.substring(0, 1)) ||
        					! newSpanString.substring(newSpanStringLen - 1, newSpanStringLen).equals(spanText.substring(spanTextLen-1, spanTextLen))
        					) {
    					if (! newSpanString.replaceAll("[\\r\\n\\s]+", " ").equals(spanText)) {
    						if (verbose) {
    							System.out.println("Case 1");
    						}
    					}
    					if (! newSpanString.substring(0,1).equals(spanText.substring(0, 1))) {
    						if (verbose) {
    							System.out.println("Case 2");
    						}
    					}
    					if (! newSpanString.substring(newSpanStringLen - 1, newSpanStringLen).equals(spanText.substring(spanTextLen-1, spanTextLen))) {
    						if (verbose) {
    							System.out.println("newSpanString.substring(newSpanStringLen - 1, newSpanStringLen) = " + newSpanString.substring(newSpanStringLen - 1, newSpanStringLen));
    							System.out.println("spanText.substring(spanTextLen-1, spanTextLen) = " + spanText.substring(spanTextLen-1, spanTextLen));
    						}
    					}
    							
    					if (verbose) {
    						System.out.println("Failed in requirement transfer");
    					}
    					return null;
    				}
        				
        		} else {
        			if (verbose) {
        				System.out.println("Something went wrong with extracting span text from DOM node, "
	        					+ "span texts are totally not equal, ignore this span");
        			}
	        		
	        		return null;
        		}
        	}
        	
        } catch (StringIndexOutOfBoundsException e) {
        	if (verbose) {
        		System.out.println("Something went wrong with extracting span text from DOM node, ignore this span");
        	}
    		return null;
        }
        
        return new LocationPosition(startOccurenceIndex, endOccurenceIndex);
        
	}

	void deleteRemovedNodes() {
    	for (Node node : debugRemovedNodes.keySet()) {
    		Node parent = node.getParentNode();
    		if (parent != null) {
    			parent.removeChild(node);
    		}
    	}
    }
    
}
