/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.documents;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.documents.IDocumentProcessor;
import com.unitesk.requality.documents.IProgressMonitor;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.ui.dialogs.SelectDocDialog;
import com.unitesk.requality.nodetypes.Document;

// TODO: headless document processing support

public abstract class DocumentProcessor implements IDocumentProcessor {
	public static boolean eclipseGUI = false; 
    protected String encoding = "UTF-8";
    protected boolean showSuperMessageBox = true;
    protected boolean showMessages = true;
    protected String messageBoxTitle = null;
    protected String messageBoxMessage = null;
    
    public void toggleShowMessages(){
    	showMessages = !showMessages;
    }
    
    @Override 
    public int[] getCounterMax() {
		return new int[]{1000,0,400};
	};
    
    @Override
    public void setEncoding(String encoding) {
        this.encoding = encoding;
        
        
    }
    int[] progress = new int[]{0,0,0};
	private IProgressMonitor monitor;
    private int sumCountersMax(){
    	int[]maxs = getCounterMax();
    	int sum = 0;
		for(int i =0;i<3;i++)
			sum+=maxs[i];
		return sum;
    }
    protected IProgressMonitor getMonitor()
    {
    	return monitor;
    }
    private int lastTargetWorked = 0;
	private void updateProgressMonitor(){
		int s = 0;
		for(int i =0;i<3;i++)
			s+=progress[i];
		if(monitor!=null)
			monitor.worked(s-lastTargetWorked);
		lastTargetWorked = s;
	}
	@Override 
	public void setCounterValues(int lc,int pc,int sc){
		progress[0] = lc;
		progress[1] = pc;
		progress[2] = sc;
		updateProgressMonitor();
	}
	@Override
	public int[]getCounterValues(){
		return progress;
	}
    @Override
    public void process(String docQId, TreeNode rootNode, IProgressMonitor monitor) throws IOException {
        //Shell parent = PlatformUI.getWorkbench().getActiveWorkbenchWindow()
          //      .getShell();

    	progress = new int[]{0,0,0};
    	lastTargetWorked = 0;
        TreeDB db = rootNode.getTreeDB();
        if(db.getStorage() instanceof TransactionStorage){
        	showErrorBox("Document Processor Cancelled", "Document Processor Cancelled: Project catalog is used by other process", null);
    		return;
    	}
 
        if(db.startTransaction("Document Processor",true)){
        	try {
        		Document doc = (Document) db.getNode(docQId);
        		String htmlName = doc.getResourceName();
        		MonitoredInputStream inStream = new MonitoredInputStream(doc.getResourceContent(htmlName),0);
        		//inStream = new BufferedInputStream(doc.getResourceContent(htmlName));
        		this.monitor = monitor;
        		Reader reader = new InputStreamReader(inStream,
                    encoding);
        		if(monitor!=null)
        			monitor.beginTask("Document '"+docQId+"' processing", sumCountersMax());
        		ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        		Writer writer = new OutputStreamWriter(byteOut, encoding);
        		if (!isCanceled()&&process(docQId, reader, writer, rootNode)) {
        			writer.close();
        			reader.close();
        			inStream.close();
        			if(isCanceled()){
        				writer.close();
        				reader.close();
        				db.rollback();
        				showMessageBox("Document Processor", "Document processing canceled");
        				return;
        			}
        			MonitoredInputStream outIn = new MonitoredInputStream(new ByteArrayInputStream(byteOut.toByteArray()),2);
        			doc.setResourceContent(htmlName, outIn, new ProcessorData());
        			outIn.close();
        			if(isCanceled()){
        				writer.close();
        				reader.close();
        				db.rollback();
        				showMessageBox("Document Processor", "Document processing canceled");
        				return;
        			}
//        			System.out.println("commit");
        			int[] cv = getCounterValues();
        			cv[2] = getCounterMax()[2];
        			setCounterValues(cv[0],cv[1],cv[2]);
        			db.commit();
        			if (showSuperMessageBox) {
        				showMessageBox("Document Processor", "Document processing finished");
        			} else {
        				if (messageBoxTitle != null && messageBoxMessage != null ) {
        					showMessageBox(messageBoxTitle, messageBoxMessage);
        				}
        			}
        		} else {
        			writer.close();
        			reader.close();
        			db.rollback();
        			if(isCanceled())
        				showMessageBox("Document Processor", "Document processing is cancelled");
        		}
        		
        	} catch (final Exception e) {
        		e.printStackTrace();
        		db.rollback();
        		showErrorBox("Document Processor Error",
        				"Document Processor Error: " + e.getLocalizedMessage(), new Status(IStatus.ERROR, RequalityPlugin.ID, 0,
        						e.getMessage(), null));
        	}
        }else{
        	showErrorBox("Document Processor Error",
    				"Some actions are already being performed", new Status(IStatus.ERROR, RequalityPlugin.ID, 0,"transaction started", null));
        }
    }
    protected boolean isCanceled(){
    	if(monitor==null)
    		return false;
    	return monitor.isCanceled();
    }
    private void setCounterValues(int[]vals)
    {
    	setCounterValues(vals[0], vals[1], vals[2]);
    }
    private class MonitoredInputStream extends BufferedInputStream{
    	int max;
    	int param;
    	public MonitoredInputStream(InputStream in,int param) {
			super(in);
			this.param = param;
			try {
				max = available();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		@Override 
		public synchronized int read(byte[] b, int off, int len) throws IOException {
    		int r = super.read(b, off, len);
    		int mx = getCounterMax()[param];
    		int[]cv = getCounterValues();
    		cv[param] = max==0?0:(mx-available()*mx/max);
    		setCounterValues(cv);
    		return r;
    	};
    }
    protected abstract boolean process(String docId, Reader reader, Writer writer,
            TreeNode rootNode) throws IOException;

    @Override
    public boolean configure(Document doc) {
    	return true;
    }
    protected void showMessageBox(final String messageBoxTitle, final String messageBoxMessage){
        if(!eclipseGUI || !showMessages){
        	System.out.println(messageBoxTitle+": "+messageBoxMessage);
        }else
        {
        	IWorkbench w = PlatformUI.getWorkbench();
        	w.getDisplay().syncExec(new Runnable() {
        		public void run() {
        			Shell activeShell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
        			MessageDialog.openInformation(activeShell, messageBoxTitle,
        					messageBoxMessage);
        		}
        	});
        }

    }
    protected void showErrorBox(final String title, final String message, final Status status){
        if(!eclipseGUI || !showMessages || !showMessages){
        	System.out.println(title+": "+message);
        }else{
        	IWorkbench w = PlatformUI.getWorkbench();
        	w.getDisplay().asyncExec(new Runnable() {
        		public void run() {
        			Shell activeShell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
        			ErrorDialog.openError(activeShell, title,
        					message, status==null?Status.OK_STATUS:status);
        		}
        	});
        }
    }
    private static Boolean lastquestionboxvalue;
    protected Boolean showQuestionBox(final String title, final String label1, final String message, final String label2) {
		lastquestionboxvalue = false;
    	if(eclipseGUI){
    		IWorkbench w = PlatformUI.getWorkbench();
    		w.getDisplay().syncExec(new Runnable() {
			
    			@Override
    			public void run() {
    				Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
    				lastquestionboxvalue = com.unitesk.requality.eclipse.ui.dialogs.BigMessageDialog.open(shell, title,
    						label1,
    						message, label2);					
    			}
    		});
    	}else{
    		//TODO ask for keypress
    		lastquestionboxvalue = false;
    		System.out.println(title+": "+label1+" "+message+" "+label2);
    	}
    	return lastquestionboxvalue;
	}
    private Document tempprefferedDoc;
    protected Document showSelectAnotherDocumentDialog(final Document doc, Document prefferedDoc) {
    	return showSelectAnotherDocumentDialog(doc,prefferedDoc,false);
    }
    protected Document showSelectAnotherDocumentDialog(final Document doc, Document prefferedDoc, boolean forcePreffered) {
    	tempprefferedDoc = null;
    	if(!eclipseGUI || !showMessages || !showMessages || forcePreffered){
    		return prefferedDoc;
    	}else{
    		IWorkbench w = PlatformUI.getWorkbench();
    		w.getDisplay().syncExec(new Runnable() {
				@Override
				public void run() {
					Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
	    			SelectDocDialog dialog = new SelectDocDialog(shell, doc.getTreeDB()) {
	    				@Override
	    				public boolean canSelect() {
	    					if (!super.canSelect()) return false;
	    					if (!selectedNode.getType().equals(Document.TYPE_NAME)) return false;
	    					if (selectedNode.getQualifiedId().equals(doc.getQualifiedId())) return false;
	    					return true;
	    				}
	    			};
	    			if (dialog.open() == SelectDocDialog.OK && dialog.getResult() instanceof Document) {
	    				tempprefferedDoc = (Document) dialog.getResult();
	    			}
				}
    		});
    	}
        return tempprefferedDoc;
	}
}
