/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.UUID;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Attr;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeComparator;
import com.unitesk.requality.documents.DeepFirstXMLDOMWalker;
import com.unitesk.requality.tools.RequalityCLI;

public class Document extends TreeNode {
    public static final String TYPE_NAME = "Document";
    private static final String ATTRIBUTE_FILE = "_file";
    private static final String TYPE_ROOT_QID = "/Documents";
    public static final String ATTR_IS_UPDATING = "_isUpdating";
    
    private List<UUID> locationCache;
    private Map<UUID,Integer> locationUUIdToIndex;
    private org.w3c.dom.Document xmlDoc;

    private boolean verbose = false;
    private freemarker.ext.dom.NodeModel modelCache = null;
    
    @Override
    protected void addChildUUID(UUID uid) {
    	super.addChildUUID(uid);
    	resetCache();
    };
    @Override
    public void removeChildUUID(UUID node) {
    	super.removeChildUUID(node);
    	if(locationCache.contains(node)){
    		locationCache.remove(node);
    		if(getUUId()!=getTreeDB().getStorage().getRootNode()&&
    	    		!isVirtual()&&!VirtualNode.getVirtualStorage(getTreeDB()).hasCloneParent(getUUId())){
    	    			TreeNode nde = VirtualNode.getVirtualTreeDB(getTreeDB()).getNode(getUUId());
    	    			((Document)nde).locationCache.remove(node);
    	        	}
    	}
    	//resetCache();
    };
    public Document() {
        super(null, null);
    }
    
    //Use it only on reports
    public freemarker.ext.dom.NodeModel getNodeModel(){
    	if(modelCache == null)
			try {
				modelCache = freemarker.ext.dom.NodeModel.parse(new File(getHTMLPath()));
			} catch (SAXException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} catch (ParserConfigurationException e) {
				e.printStackTrace();
			}
		return modelCache;
    }

    private Element getElementChildIgnoreCase(Node node, String name) {
        for (int i = 0; i < node.getChildNodes().getLength(); i++) {
            if (node.getChildNodes().item(i).getNodeName().toLowerCase()
                    .equals(name.toLowerCase())) {
                return (Element) node.getChildNodes().item(i);
            }
        }
        return null;
    }

    public void setMetaCharset(String charset) {
        try {
            org.w3c.dom.Document dom = getDOM();
            Node html = dom.getFirstChild();
            Node head = getElementChildIgnoreCase(dom.getFirstChild(), "head");
            if (head == null) {
                head = dom.createElement("head");
                html.appendChild(head);
            }
            Element meta = getElementChildIgnoreCase(head, "meta");
            if (meta == null) {
                meta = dom.createElement("meta");
                head.appendChild(meta);
            }
            meta.setAttribute("http-equiv", "Content-Type");
            meta.setAttribute("content", "text/html; charset=" + charset);
            saveDOM(null,dom);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public String getMetaCharset() {
        // TODO: check rfc for this meta
        String result = null;
        try {
            DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(getDOM()
                    .getFirstChild());
            for (Node node : dfw) {
                if (node.getNodeName().toLowerCase().equals("meta")) {
                    if (node.getAttributes().getNamedItem("http-equiv") == null)
                        continue;
                    if (!node.getAttributes().getNamedItem("http-equiv")
                            .getNodeValue().toLowerCase()
                            .equals("content-type"))
                        continue;
                    if (node.getAttributes().getNamedItem("content") == null)
                        continue;
                    String content = node.getAttributes()
                            .getNamedItem("content").getNodeValue();
                    int i = content.indexOf("charset=");
                    if (i < 0)
                        return null;
                    return content.substring(i + "charset=".length());
                }
            }
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return result;
    }

    public static String getTypeRootQId() {
        return TYPE_ROOT_QID;
    }

    public Document(TreeDB db, UUID uuid) {
        super(db, uuid);
        setType(TYPE_NAME);
    }

    public void setResourceName(String name) {
        Attribute a = new Attribute(this, AttributeType.STRING, ATTRIBUTE_FILE, name);
        putAttribute(a);
    }

    @Override
    public void setParent(TreeNode parent) {
        String oldDocQId = getQualifiedId();
        super.setParent(parent);
        String newDocQId = parent.getQualifiedId() + "/" + id;
        List<UUID> locs = getLocations();
        for (UUID loc : locs) {
            String oldQid = oldDocQId + "/" + loc.toString();
            String newQid = newDocQId + "/" + loc.toString();
            Requirement reqs[] = Location.getRequirementsFor(getTreeDB(),
                    oldQid);
            for (Requirement req : reqs) {
                req.removeLocation(oldQid);
                req.addLocation(newQid);
                req.saveAttributes();
            }
        }
    }

    public String getResourceName() {
        return (String) getAttributeValue(ATTRIBUTE_FILE);
    }

    // private static final String LOCATION_CLASS_ID_PREFIX =
    // "requality_text id_";
    private static final String LOCATION_CLASS_ID_PREFIX = "id_";

    public String getHTMLPath() {
        return getTreeDB().getStorage().getNodeResourcePath(getUUId())
                + File.separator + getResourceName();
    }

    public Map<UUID, Integer> getLocationIndexMap(){
    	if(locationUUIdToIndex == null)
    		getLocations();
    	return locationUUIdToIndex;
    }
    //TODO refactor
    public List<UUID> getLocations() {
        if (locationCache == null || locationUUIdToIndex == null) {
            locationCache = new ArrayList<UUID>();
            locationUUIdToIndex = new HashMap<UUID, Integer>();
            try {
            	int cnt = 0;
                DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(loadDOM()
                        .getFirstChild());
                for (Node node : dfw) {
                    //try to handle not only location with span
                    //but every that have mark in class
                //    if (node.getNodeName().equals("span")) {
                        if(node.getAttributes() == null)
                            continue;
                        Node spanClass = node.getAttributes().getNamedItem(
                                "class");
                        if (spanClass != null) {
                            String classNames[] = spanClass.getNodeValue()
                                    .split(" ");
                            for (String className : classNames) {
                                if (className
                                        .startsWith(LOCATION_CLASS_ID_PREFIX)) {
                                	String id = className
                                            .substring(LOCATION_CLASS_ID_PREFIX
                                                    .length());
                                	UUID uuid = UUID.fromString(id);
                                	try{
                                		TreeNode lnode = findChild(id);
                                		if(lnode==null)
                                			lnode = getTreeDB().getNode(uuid);
                                		if(lnode!=null)
                                			uuid = lnode.getUUId();
                                	}
                                	catch(RuntimeException re){
                                	}
                                	if (!getLocationIndexMap().containsKey(uuid)) {
                            			locationCache.add(uuid);
                            			getLocationIndexMap().put(uuid, cnt++);
                            		}
                                }
                            }
                        }
                  //  }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        if(locationCache == null)
        	return Collections.EMPTY_LIST;
        return locationCache;
    }

    public int getLocationIndex(UUID locuid){
    	if(locationUUIdToIndex == null)
    		getLocations();
    	if(locationUUIdToIndex == null)
    		return -1;
    	if(locationUUIdToIndex.containsKey(locuid))
    		return locationUUIdToIndex.get(locuid);
    	int num = sizeChildren();
        for (int i = 0; i < num; i++) {
            TreeNode child = getChild(i);
            if (child.getUUId().equals(locuid)) {
            	UUID id = UUID.fromString(child.getId());
            	if(getLocationIndexMap().containsKey(id))
            		return getLocationIndexMap().get(id);
            }
        }
    	return -1;
    }
    
    public static class LocationRegion {
        public LocationRegion() {
            startNodeType = endNodeType = Node.ELEMENT_NODE;
        }

        public LocationRegion(String sX, int sT, int sOff, String eX, int eT, int eOff, String text) {
            startNodeXPath = sX;
            startNodeType = sT;
            startNodeOffset = sOff;
            endNodeXPath = eX;
            endNodeType = eT;
            endNodeOffset = eOff;
            this.text = text;
        }

        public boolean isStarted() {
            return startNodeXPath != null;
        }

        private String startNodeXPath;
        private int startNodeType;
        private int startNodeOffset;

        private String endNodeXPath;
        private int endNodeType;
        private int endNodeOffset;

        private String text = "";

        @Override
        public String toString() {
            String start = makeNodeString("start", startNodeXPath,
                    startNodeType, startNodeOffset);
            if(endNodeXPath==null) {
                endNodeXPath = startNodeXPath;
                endNodeOffset=startNodeOffset + text.length();
            }
            String end = makeNodeString("end", endNodeXPath, endNodeType,
                    endNodeOffset);
            return "{" + start + "," + end + ",\"text\":\"" + text + "\"}";
        }

        private String makeNodeString(String name, String xpath, int type,
                int offset) {
            StringBuilder sb = new StringBuilder();
            sb.append("\"");
            sb.append(name);
            sb.append("\":{\"node\":\"");
            sb.append(xpath);
            sb.append("\",\"nodeType\":");
            sb.append(type);
            sb.append(",\"offset\":");
            sb.append(offset);
            sb.append("}");
            return sb.toString();
        }

        public void text(String text) {
            if(!isText()) {
                startNodeType = endNodeType = Node.TEXT_NODE;
            }
            this.text+=text;
        }
        public boolean isText() {
            return startNodeType == Node.TEXT_NODE;
        }

        public void xpath(String xpath) {
            if(!isStarted()) {
                startNodeXPath = xpath;
            } else {
                endNodeXPath = xpath;
            }
        }

        public void offset(int offset) {
            if(!isStarted()) {
                startNodeOffset = offset;
            } else {
                endNodeOffset = offset;
            }
        }
    }

    public String removeLocation(String uuid) {
        try {
        	//resetCache();
            List<LocationRegion> regions = new ArrayList<LocationRegion>();
            removeLocation(UUID.fromString(uuid), getDOM().getFirstChild(),
                    regions, 0);

            StringBuilder sb = new StringBuilder();
            sb.append("{\"action\":\"delete\",\"uuid\":\"");
            sb.append(uuid);
            sb.append("\",\"selection\":[");
            boolean first = true;
            
            //for (int i = regions.size()-1;i>=0;i--) {
        	//LocationRegion r = regions.get(i);
            for (LocationRegion r : regions) {
                if (!first) 
                    sb.append(",");
                sb.append(r.toString());
                first=false;
            }
            sb.append("]}");
            String result = sb.toString(); 
            saveDOM(result);
            if(locationCache!=null && locationCache.contains(uuid)){
        		locationCache.remove(uuid);
        		if(getUUId()!=getTreeDB().getStorage().getRootNode()&&
        	    		!isVirtual()&&!VirtualNode.getVirtualStorage(getTreeDB()).hasCloneParent(getUUId())){
        	    			TreeNode nde = VirtualNode.getVirtualTreeDB(getTreeDB()).getNode(getUUId());
        	    			((Document)nde).locationCache.remove(uuid);
        	        	}
        	}
            saveDOM(null,xmlDoc);
            return result;
        } catch (Exception e) {
            e.printStackTrace();
        }
        saveDOM(null,xmlDoc);
        return null;
    }

    private static boolean needDropNode(Node node, UUID uuid) {
        if (node.getNodeName().equals("a")) {
            if (node.getAttributes().getNamedItem("id").getNodeValue()
                    .equals("id_" + uuid)) {
                return true;
            }
        }

        if (node.getNodeName().equals("span")) {
            String className = node.getAttributes().getNamedItem("class")
                    .getNodeValue();
            if (className.startsWith("requality_link requality_link_" + uuid)) {
                return true;
            }
            if (className
                    .startsWith("requality_link requality_link_in_" + uuid)) {
                return true;
            }
        }

        return false;
    }

    public Location createLocation() {
    	return createLocation(null);
    }

    public Location createLocation(UUID id) {
    	if(id == null)
    		id = UUID.randomUUID();
        Location l = (Location) getTreeDB().createChildNode(id, getUUId(), id.toString(), Location.TYPE_NAME);
        // saveDOM(); // moved to ReqMarker.addSelectionTo  
        return l;
    }

    // TODO: nested location support
    private void removeLocation(UUID uuid, Node rnode,
            List<LocationRegion> regions, final int parentOffset) {
    	Map<Node,int[]>later = new HashMap<Node,int[]>();
        Queue<Node> q = new LinkedList<Node>();
        Queue<Node> oldpar = new LinkedList<Node>();
        q.add(rnode);
        List<Node> selNodes = new LinkedList<Node>();
    	Map<Node,Integer> selTexts = new HashMap<Node, Integer>();
    	
        while(!q.isEmpty()){
        	Node node = q.poll();
	        if (isLocationNode(node)){ 
	            if (getUUID(node).equals(uuid)) {
	            	if(!oldpar.contains(node.getParentNode()))
	            		oldpar.add(node.getParentNode());
	            	Node parNode = node.getParentNode();
	            	NodeList nl = node.getChildNodes();
	            	if(nl!=null)
	    				for(int i = 0;i<nl.getLength();i++){
	    					Node nde = nl.item(i);
	    					if (needDropNode(nde, uuid)) 
		                        continue;
	    					if(nde.getNodeType()==Node.TEXT_NODE){
	    						if(!selTexts.keySet().contains(nde))
	    							if(nde.getNodeValue().replace("\n", "").length()>0)
	    								selTexts.put(nde, nde.getNodeValue().length());
	    					}
	    					else
	    						if(selNodes.indexOf(nde)<0){
	    							String s = nde.getAttributes().getNamedItem("class").getTextContent().replaceAll(" id_"+uuid.toString(), "");
	    							if(s.trim().equals("requality_node"))
	    								nde.getAttributes().removeNamedItem("class");
	    							else{
	    								 Attr cl = nde.getOwnerDocument().createAttribute("class");
	    								 cl.setValue(s);
	    								 nde.getAttributes().setNamedItem(cl);
	    							}
    								later.put(nde, new int[]{-1,-1});
	    						}
	    				}
	    			for (int i = 0; i < node.getChildNodes().getLength(); i++) {
	                    Node curNode = node.getChildNodes().item(i);
	                    if (needDropNode(curNode, uuid)) {
	                        continue;
	                    }
	
	                    //if(curNode.getNodeType()==Node.TEXT_NODE) {
	                      //  len = curNode.getTextContent().length();
	//                        r.text(curNode.getTextContent());
	  //                  }
						Node nd = node.getNextSibling(); 
						if(nd!=null)
	                    	parNode.insertBefore(curNode,
	                            	nd);
	                   	else
	                   		 parNode.appendChild(curNode);
	                }
	                parNode.removeChild(node);
		        }
	        }else{
	        	NodeList cnl = node.getChildNodes();
	        	for(int i = 0;i<cnl.getLength();i++)
		        	q.add(cnl.item(i));
	        }
        }
        for(Node parNode:oldpar){
        	NodeList nl = parNode.getChildNodes();
			for(int i =0;i<nl.getLength();i++){
				int j = i;
				Node nodeNow=nl.item(j);
				if(nodeNow.getNodeType()==Node.TEXT_NODE){
					int len = -1;
					Node b =null;
					Iterator<Node>it=selTexts.keySet().iterator();
					for(int k=0;k<selTexts.keySet().size();k++)
					{
						Node knl = it.next();
						if(knl.isEqualNode(nodeNow)){
							b = knl;
							len = selTexts.get(knl);
						}
					}
	    			boolean processed = false;
	    			Node next = nodeNow.getNextSibling();
					int start = -1;
					int end = -1;
					if(b!=null){
						end = len;
						start = 0;
					}
					while (next!=null
						&& next.getNodeType() == Node.TEXT_NODE) {
						Node nextFound = null;
						Iterator<Node>it2=selTexts.keySet().iterator();
						for(int k=0;k<selTexts.keySet().size();k++){
							Node n =it2.next(); 
							if(n.isEqualNode(next))
								nextFound=n;
						}
						if(nextFound!=null){
							processed = true;
							if(end == -1)
								end = nodeNow.getNodeValue().length(); 
							end += selTexts.get(nextFound);
							if(start == -1)
								start = nodeNow.getNodeValue().length();
						}
						nodeNow.setNodeValue(nodeNow.getNodeValue()+next.getNodeValue());
						parNode.removeChild(next);
						next = nodeNow.getNextSibling();
					}
					if(processed||b!=null){
						later.put(nodeNow, new int[]{start,end});
					}
	    			//if(b!=null)
	    				//selTexts.remove(b);
    			}
    		}
        }
        rnode.getOwnerDocument().normalize();
        for(Node n: later.keySet()){
        	int[] inf = later.get(n);
			LocationRegion r = new LocationRegion();
			Object[]o=getElementXpath(n);
			r.startNodeXPath = r.endNodeXPath = (String) o[0];
			r.startNodeType = r.endNodeType = (Short) o[1];
			if(inf[0]>-1){
				r.startNodeOffset = inf[0];
				r.endNodeOffset = inf[1];
			}
			regions.add(r);
        }
        saveDOM(null,rnode.getOwnerDocument());
    }

    private LocationRegion currentRegion(List<LocationRegion> regions) {
        if (regions.size() == 0) {
            regions.add(new LocationRegion());
        }
        return regions.get(regions.size() - 1);
    }

    public void saveDOM() {
        saveDOM(null);
    }
    public void saveDOM(String jsData) {
    	saveDOM(jsData,null);
    }
    private void saveDOM(String jsData, org.w3c.dom.Document doc) {
    	if(doc == null)
    		try {
    			doc = getDOM();
    		} catch (Exception e) {
    			e.printStackTrace();
    		} 
            
        long t_start = System.currentTimeMillis();
        try {
            ByteArrayOutputStream newXml = new ByteArrayOutputStream();
            Writer out = new OutputStreamWriter(newXml, "UTF-8");
            Result result = new StreamResult(out);
            TransformerFactory.newInstance().newTransformer()
                    .transform(new DOMSource(doc), result);
            out.close();
            setResourceContent(getResourceName(), new ByteArrayInputStream(
                    newXml.toByteArray()), jsData);

        } catch (Exception e) {
            e.printStackTrace();
        }
        // TODO: optimize cache
        resetCache();
        
        if (verbose) {
        	long t_end = System.currentTimeMillis();
        	System.out.println("time Document.saveDOM(" + jsData + ") = " + (t_end - t_start));
        }
    }
    
    public void saveDOMFromXMLCache() {
    	long t_start = System.currentTimeMillis();
        ByteArrayOutputStream newXml = new ByteArrayOutputStream();
        try {
        	org.w3c.dom.Document doc = getXmlDoc();
            Writer out = new OutputStreamWriter(newXml, "UTF-8");
            Result result = new StreamResult(out);
            TransformerFactory.newInstance().newTransformer()
                    .transform(new DOMSource(doc), result);
            out.close();
            setResourceContent(getResourceName(), new ByteArrayInputStream(
                    newXml.toByteArray()), null);

        } catch (Exception e) {
            e.printStackTrace();
        }
        // TODO: optimize cache
        resetCache();
        
        
        if (verbose) {
        	long t_end = System.currentTimeMillis();
        	System.out.println("time Document.saveDOMFromXMLCache() = " + (t_end - t_start));
        }
    }
    

    private boolean isLocationNode(Node node) {
        if (node.getNodeName().equals("span")) {
            Node spanClass = node.getAttributes().getNamedItem("class");
            if (spanClass == null) {
                return false;
            }
            String className = spanClass.getNodeValue();
            if (className.contains(LOCATION_CLASS_ID_PREFIX)
                    && className.contains("requality_text")) {
                return true;
            }
        }
        return false;
    }

    private UUID getUUID(Node locationNode) {
        try {
            String classNames[] = locationNode.getAttributes()
                    .getNamedItem("class").getNodeValue().split(" ");
            for (String className : classNames) {
                if (className.startsWith(LOCATION_CLASS_ID_PREFIX)) {
                    return UUID.fromString(className
                            .substring(LOCATION_CLASS_ID_PREFIX.length()));
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        return null;
    }

    public Location getLocation(String locationId) {
    	Location l = (Location) findChild(locationId);
    	if(l == null)
    		try {
				l = (Location) getTreeDB().getNode(UUID.fromString(locationId));
			} catch (Exception e) {
			}
    	return l;
    }

    public org.w3c.dom.Document getDOM() throws Exception {
//        if (getXmlDoc() == null) {
//            setXmlDoc(loadDOM());
//        }
//    	setXmlDoc(loadDOM());
//        return getXmlDoc();
    	
    	setXmlDoc(loadDOM());
    	return getXmlDoc();
    }

    public org.w3c.dom.Document loadDOM() throws Exception {
    	InputStream in = getResourceContent(getResourceName());
        
        long t_start = System.currentTimeMillis();
        DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance()
                .newDocumentBuilder();
        org.w3c.dom.Document result = docBuilder.parse(in);
        in.close();
        result.normalize();
        
        if (verbose ) {
        	long t_end = System.currentTimeMillis();
        	System.out.println("time Document.loadDOM() = " + (t_end - t_start));
        }
        
        return result;
    }

    private boolean containsLocationId(String[] classNames) {
        boolean requalityTextClass = false;
        boolean locationIdClass = false;

        for (String name : classNames) {
            if (name.equals("requality_text")) {
                requalityTextClass = true;
            }
            if (name.startsWith(LOCATION_CLASS_ID_PREFIX)) {
                locationIdClass = true;
            }
            if (requalityTextClass && locationIdClass) {
                return true;
            }
        }
        return requalityTextClass && locationIdClass;
    }

    private String getLocationIdClass(String[] classNames) {
        for (String name : classNames) {
            if (name.startsWith(LOCATION_CLASS_ID_PREFIX)) {
                return name;
            }
        }
        return null;
    }

    public String extractLocationText(String locUUID) {
        try {

        	org.w3c.dom.Document d = getDOM();
            DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(d
                    .getFirstChild());

            DocumentFragment fragment = d.createDocumentFragment();
            for (Node node : dfw) {
                if (node.getNodeName().equals("span")) {
                    Node spanClass = node.getAttributes().getNamedItem("class");
                    if (spanClass != null) {
                        String[] classNames = spanClass.getNodeValue().split(
                                " ");

                        if (containsLocationId(classNames)) {
                            String locationIdClass = getLocationIdClass(classNames);
                            UUID uuid = UUID.fromString(locationIdClass
                                    .substring(LOCATION_CLASS_ID_PREFIX
                                            .length()));
                            if (uuid.toString().equals(locUUID)) {
                                for (int i = 0; i < node.getChildNodes()
                                        .getLength(); i++) {
                                    Node ch = node.getChildNodes().item(i);
                                    if (ch.getNodeName().equalsIgnoreCase("a")
                                            && ch.getAttributes().getNamedItem(
                                                    "name") != null
                                            && ch.getAttributes()
                                                    .getNamedItem("name")
                                                    .getNodeValue()
                                                    .equals(locUUID)) {
                                        continue;
                                    }
                                    fragment.appendChild(ch.cloneNode(true));
                                }
                            }
                        }
                    }
                }
            }
            ByteArrayOutputStream result = new ByteArrayOutputStream();
            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
                    "yes");
            transformer.transform(new DOMSource(fragment), new StreamResult(
                    result));
            result.flush();
            return result.toString("UTF-8");

        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    public int sizeLocations() {
        return getLocations().size();
    }

    @Override
    public String toString() {
        return getId();
    }

    public void resetCache() {
        // TODO: optimize cache
    	if (verbose) {
    		System.out.println("in resetCache");
    	}
        locationCache = null;
        locationUUIdToIndex = null;
        modelCache = null;
        if(getUUId()!=getTreeDB().getStorage().getRootNode()&&
    		!isVirtual()&&!VirtualNode.getVirtualStorage(getTreeDB()).hasCloneParent(getUUId())){
    			TreeNode node = VirtualNode.getVirtualTreeDB(getTreeDB()).getNode(getUUId());
    			if (node.isVirtual())
    				((Document)node).resetCache();
        	}
    }

    protected org.w3c.dom.Document getXmlDoc() {
        return xmlDoc;
    }

    public void setXmlDoc(org.w3c.dom.Document xmlDoc) {
        this.xmlDoc = xmlDoc;
    }

    public String createLocationRangesDescriptor(String locId) {
        try {
            getDOM();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return "";
    }

    // /////////////////////////
    // xpath stuff

    private static Set<Node> recLook(Node n){
    	Set<Node> list = new HashSet<Node>();
    	NodeList nl = n.getChildNodes();
    	
    	for(int i=0;i<nl.getLength();i++)
    		if(!list.contains(nl.item(i))){
    			list.add(nl.item(i));
    			list.addAll(recLook(nl.item(i)));
    		}
    	return list;
    }
    public static Object[] getElementXpath(Node nodeNow) {
    	Set<Node> allnodes = recLook(nodeNow.getOwnerDocument());
    	Node[]aallNodes = allnodes.toArray(new Node[0]);
        String path = "";
        short type = Node.ELEMENT_NODE;
        if(nodeNow.getNodeType()==Node.TEXT_NODE)
        {
        	type = Node.TEXT_NODE;
        	nodeNow = nodeNow.getParentNode();
        }
        try {
            for (; nodeNow != null && nodeNow.getNodeType()==Node.ELEMENT_NODE; nodeNow = nodeNow.getParentNode()) {
            	if(((Element)nodeNow).hasAttribute("id")){
            		short uniqueIdCount = 0;
                	for(int i = 0;i<aallNodes.length;i++){
                		if (aallNodes[i].getAttributes()!=null && aallNodes[i].getAttributes().getNamedItem("id")!=null && aallNodes[i].getAttributes().getNamedItem("id").getNodeValue().equals(nodeNow.getAttributes().getNamedItem("id").getNodeValue()))
        					uniqueIdCount++;
        				if (uniqueIdCount > 1)
        					break;
                	}
                	if(uniqueIdCount == 1){
                		path = "id(\""+nodeNow.getAttributes().getNamedItem("id").getNodeValue()+"\")" +"/" + path;
                       	path = path.substring(0,path.length()-1);
                    	return new Object[]{path,type};
                	}else
                		path = "x:"+nodeNow.getLocalName().toLowerCase()+"[@id=\""+
                				nodeNow.getAttributes().getNamedItem("id").getNodeValue()+"\"]"+"/"+ path;
            	}else
            	{
            		int i;
            		Node sib;
            		for(i = 1, sib = nodeNow.getPreviousSibling();sib!=null;sib=sib.getPreviousSibling()){
            			if(sib.getNodeName()!=null && sib.getNodeName().equals(nodeNow.getNodeName()))
            				i++;
            		}
            		path = "x:"+nodeNow.getNodeName().toLowerCase()+"["+i+"]"+"/"+path;
            	}
            }
            if(path.length() == 0)
            	return null;
            else
            	path = path.substring(0,path.length()-1);
        	return new Object[]{path,type};
        } catch (Exception ee) {
        	ee.printStackTrace();
        }
        return null;
    }
   
    public void duplicateImages(String path){
    	RequalityCLI.duplicateImages(this, path);
    }

	public int compareTo(Document document) {
		String qid1=getQualifiedId(),qid2=getQualifiedId();
		int i = 0, li = 0;
		while(i>=0 && qid1.subSequence(0, i).equals(qid2.subSequence(0, i)) && i<qid2.length()){
			li = i;
			i = qid1.indexOf("/", i+1);
		}
		int id1=qid1.indexOf("/", li+1), id2=qid2.indexOf("/", li+1);
		if(id1==-1)id1 = qid1.length();
		if(id2==-1)id2 = qid2.length();
		TreeNode n1 = getTreeDB().getNode(qid1.substring(0, id1)),
				 n2 = getTreeDB().getNode(qid2.substring(0, id2));
		return TreeNodeComparator.getComparator().compare(n1, n2);
	}
	
	public void addUpdatedProblem() {
		putAttribute(new Attribute(this, AttributeType.STRING, ATTR_IS_UPDATING, "updating"));
		saveAttributes();
	}

}
