/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.tools;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.FSStorage;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.documents.DocumentImporter;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.core.ResourceStorage;
import com.unitesk.requality.eclipse.core.TreesTracker;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.report.ReportGenerator;
import com.unitesk.requality.report.Template;

//
public class RequalityCLI {

    public final static String cmdName = "requality";
    private static Map<String, String> usages = new HashMap<String, String>();

	public static String getSemiAbsolutePath(String filename, TreeDB db) {
		if(db.isVirtual()&&VirtualNode.getBaseTree(db)!=null)
			db = VirtualNode.getBaseTree(db);
		ITreeStorage ts = db.getStorage();
		if(ts instanceof ResourceStorage)
			return TreesTracker.getTracker().getProject(db).getLocation().toOSString()+"/"+filename;
		if(ts instanceof FSStorage)
			return ((FSStorage)ts).getResourceFolder(ts.getRootNode()).getParent()+"/"+filename;
		return filename;
	}
    static {

        usages.put(
                "report",
                "Usage: "
                        + cmdName
                        + " report <template> <catalogue> <output folder> [-root=\"<root qid>\"] [-<param>=\"<value>\"]\n"
        				+ "Here: \n"
                        + "<template> - a path to a report template (.xml file). All templates delivered with Requality plugin are placed in \"templates\" subfolder.\n"
                        + "<catalogue> - a path to the Requality project to be used as a source for report generation.\n"
                        + "<output folder> - a path to the folder where report to be generated.\n"
                        + "<root qid> - qualified identifier of requirement to be used as a root for report generation. Default value - \"/Requirements\".\n"
                        + "[-<param>=\"<value>\"] - a set of report-specific parameters. See documentation for the corresponding report template if it supports additional parameters.");
        usages.put(
                "convert",
                "Usage: "
                        + cmdName
                        + " convert "
                        + "<path to database> <path to document> [<document folder qid>] [<group requirement qid>]");
    }

    public static void usage() {
        log("Usage: " + cmdName + " " + "<help|convert|report> [params]");
    }

    public static boolean deleteDirectory(File directory) {
        if(directory.exists()){
            File[] files = directory.listFiles();
            if(null!=files){
                for(int i=0; i<files.length; i++) {
                    if(files[i].isDirectory()) {
                        deleteDirectory(files[i]);
                    }
                    else {
                        files[i].delete();
                    }
                }
            }
        }
        return(directory.delete());
    }
    
    public static String getFileContent(File f){
		StringBuffer fileData = new StringBuffer();
        try {
        	BufferedReader reader = new BufferedReader(
                new FileReader(f));
        	char[] buf = new char[1024];
        	int numRead=0;
        		while((numRead=reader.read(buf)) != -1){
					String readData = String.valueOf(buf, 0, numRead);
			    	fileData.append(readData);
				}
			reader.close();
        } catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        return fileData.toString();
	}
    
    public static void usageHelp(String param) {
        StringBuilder result = new StringBuilder();

        if (usages.containsKey(param)) {
            result.append(usages.get(param));
        } else {
            usage();
        }

        log(result.toString());
    }

    private static void log(String string) {
        System.out.println(string);
    }

    public static void main(String[] args) throws Exception {
        if (args.length == 0) {
            usage();
            return;
        }
        System.out
                .println("Welcome to Requality console mode! Process started...");

        if ("help".equals(args[0])) {
            String subj = args.length > 1 ? args[1] : "";
            usageHelp(subj);
            return;
        }
        // [<document folder qid>] [<group requirement qid>]
        if ("convert".equals(args[0])) {
            if (args.length > 2 && args.length < 6) {
                String pathToDB = args[1];
                String pathToDoc = args[2];
                String[] qids = new String[args.length - 3];
                for (int i = 3; i < args.length; i++) {
                    qids[i - 3] = args[i];
                }

                File inFile = new File(pathToDoc);
                TreeDB db = TreeDB.load(new FSStorage(pathToDB));
                db.addNodeType(Requirement.TYPE_NAME, Requirement.class);
                db.addNodeType(Location.TYPE_NAME, Location.class);
                db.addNodeType(DocFolder.TYPE_NAME, DocFolder.class);
                db.addNodeType(Document.TYPE_NAME, Document.class);
                db.addNodeType(VirtualNode.TYPE_NAME,VirtualNode.class);
                Requirement groupNode = (Requirement) db.getNode(Requirement
                        .getTypeRootQId());
                DocFolder folderNode = (DocFolder) db.getNode(Document
                        .getTypeRootQId());
                if(groupNode == null)
                {
                	System.out.println("target catalog cannot be loaded or target node does not exist");
                	return;
                }
                
                db.startTransaction("convert");
                Document doc = (Document) folderNode.createChild(
                        inFile.getName(), Document.TYPE_NAME);
                String name = inFile.getName();
                int ind = name.lastIndexOf(".");
                if(ind>-1)
                	name = name.substring(0,ind);
                doc.setResourceName(name+".xhtml");
                doc.saveAttributes();
                db.commit();
                Converter c = new Converter();
                try {
                	File res = new File(doc.getHTMLPath()).getParentFile();
                	if(!res.exists())
                		res.mkdirs();

                    Writer out = new FileWriter(doc.getHTMLPath());
                    StringWriter tmpOut = new StringWriter();
                    DocumentImporter.cleanHtml(new FileReader(
                            inFile), tmpOut);
                    Map<String, String[]> result = c.convert(new StringReader(tmpOut.toString()), out);
                    c.store(result, doc.getQualifiedId(), db, groupNode);
                    System.out.println("OK");
                } catch (IOException e) {
                    System.out.println("ERROR");
                    e.printStackTrace();
                }
                return;
            }
            usageHelp("convert");
            return;
        }

        if ("report".equals(args[0])) {
            if (args.length < 4) {
                usageHelp("report");
                return;
            }
            String reportArgs[] = Arrays.copyOfRange(args, 1, args.length);
            TreeDB db = TreeDB.load(new FSStorage(reportArgs[1]));
            db.addNodeType(Requirement.TYPE_NAME, Requirement.class);
            db.addNodeType(Location.TYPE_NAME, Location.class);
            db.addNodeType(DocFolder.TYPE_NAME, DocFolder.class);
            db.addNodeType(Document.TYPE_NAME, Document.class);
            db.addNodeType(TestPurpose.TYPE_NAME, TestPurpose.class);
            db.addNodeType(VirtualNode.TYPE_NAME,VirtualNode.class);

            String templatePath = reportArgs[0];
            Template tpl = new Template(templatePath);
            String rootQId = Requirement.getTypeRootQId();
            String saveTo = reportArgs[2];
            
            TreeDB instTree = VirtualNode.getVirtualTreeDB(db);
            
            instTree.addNodeType(Requirement.TYPE_NAME, Requirement.class);
            instTree.addNodeType(Location.TYPE_NAME, Location.class);
            instTree.addNodeType(DocFolder.TYPE_NAME, DocFolder.class);
            instTree.addNodeType(Document.TYPE_NAME, Document.class);
            instTree.addNodeType(TestPurpose.TYPE_NAME, TestPurpose.class);
            instTree.addNodeType(VirtualNode.TYPE_NAME,VirtualNode.class);

            Map<String,Object> params = new HashMap<String, Object>();
            
            for(int i = 3; i < reportArgs.length; i++)
            	if (reportArgs[i].startsWith("-")) {
            		int k = reportArgs[i].indexOf("=");
            		if (k < 0){
            			log("user-defined parameter '"+reportArgs[i]+"' is not provided with value in form of param=\"value\"");
            			continue;
            		}
            		String name = reportArgs[i].substring(1,k), value = reportArgs[i].substring(k+1);
            		if(value.startsWith("\"")&&value.endsWith("\""))
            			value = value.substring(1,value.length()-1);
            		if(name.equals("root"))
            			rootQId = value;
            		else
            			params.put(name, value);
            	}
            Requirement rootReq = null;
            try{
            	rootReq = (Requirement) instTree.getNode(rootQId);
            	if(rootReq == null){
                	rootReq = (Requirement) instTree.getNode(Requirement.getTypeRootQId());
                	log("target id is not valid, default value is used");
            	}
            }catch(Exception e){
            	log("root id is not valid or catalog cannot be loaded");
            	rootReq = (Requirement) instTree.getNode(Requirement.getTypeRootQId());
            }
            
            ReportGenerator.generateReport(rootReq, tpl, saveTo,
                    params);
            return;
        }
        // if ("docclean".equals(args[0])) {
        // if (args.length == 2) {
        // docclean(args[1]);
        // } else {
        // System.out.println("neep path to database");
        // }
        // return;
        // }

        usage();
    }

    // private static void docclean(String path) {
    // TreeDB db = new TreeDBIO(path);
    // Map<String, Set<UUID>> docToUUIDs = new HashMap<String, Set<UUID>>();
    // for (TreeDBNode r : new DeepFirstTreeWalker(db.getRootNode(
    // Requirement.type, false))) {
    // Requirement req = (Requirement) r;
    // for (Location loc : req.getLocations()) {
    // String docId = loc.getDocId().replace('\\', '/');
    // if (docToUUIDs.get(docId) == null) {
    //
    // docToUUIDs.put(docId, new HashSet<UUID>());
    // }
    // docToUUIDs.get(docId).add(loc.getId());
    // }
    // }
    // for (String docId : docToUUIDs.keySet()) {
    // Set<UUID> locRemove = new HashSet<UUID>();
    // Document doc = db.getDocument(docId);
    // System.out.println("processing document: " + docId + " locations ("
    // + docToUUIDs.get(docId).size() + "/" + doc.sizeLocations()
    // + ")");
    //
    // for (UUID locId : doc.getLocations()) {
    // locRemove.add(locId);
    // }
    // locRemove.removeAll(docToUUIDs.get(docId));
    // for (UUID locId : locRemove) {
    // System.out.println("Remove location " + locId.toString()
    // + " from " + docId);
    // doc.removeLocation(locId);
    // }
    // }
    // }
    public static File copyfile(String srFile, String dtFile)
            throws IOException {
    	return copyfile(null, srFile, dtFile);
    }
    public static File copyfile(TreeNode node, String srFile, String dtFile)
            throws IOException {
    	return copyfile(node, srFile, dtFile, false);
    }
    
    public static File copyfile(TreeNode node, String srFile, String dtFile, boolean rewrite)
            throws IOException {
        File sourceFile = new File(srFile);
        File destFile = new File(dtFile);
        boolean localfile = true;
    	if (!sourceFile.exists()) {
        	localfile = false;
          //  throw new IOException(sourceFile+" not exists");
        }
        ReadableByteChannel source = null;
        if (!localfile){
        	try{
        		URL website = new URL(srFile);
        		source = Channels.newChannel(website.openStream());
        		if(!source.isOpen())
        			throw new IOException(sourceFile + " not exists");
        	}
        	catch(MalformedURLException me){
        		throw new IOException(sourceFile + " not exists");
        	}
        	catch(IOException ie){
        		throw new IOException(sourceFile + " not exists");
        	}
        }
        File possibleParent = destFile.getParentFile();
        if(!destFile.getParentFile().exists())
        {
        	//hack to not to break resource system
        	if(node != null){
        		destFile = new File(possibleParent.getParent()+File.separator+destFile.getName());
        	}else{
        		possibleParent.mkdirs();
        	}
        }
        if (!destFile.exists()) {
            destFile.createNewFile();
        }else
        	if(!rewrite){
        		return destFile;
        	}else{
        		destFile.delete();
        		destFile.createNewFile();
        	}
        FileChannel destination = null;
        try {
        	if (localfile)
        		source = new FileInputStream(sourceFile).getChannel();
            destination = new FileOutputStream(destFile).getChannel();
            destination.transferFrom(source, 0, Long.MAX_VALUE);
        } finally {
            if (source != null) {
                source.close();
            }
            if (destination != null) {
                destination.close();
            }
        }
        if(node != null){
        	String s = destFile.getName();
        	File temp = new File(destFile.getAbsolutePath()+"_temp");
        	destFile.renameTo(temp);
        	if(node.isResourceExists(s))
        		node.deleteResource(s);
        	node.setResourceContent(s, new FileInputStream(temp));
        	temp.delete();
        	destFile = new File(possibleParent.getAbsolutePath()+File.separator+s);
        }
        return destFile;
    }

    public static String getNumberStringFromId(String in) {
        Pattern pattern = Pattern.compile("(.*\\D|)(\\d*)$");
        Matcher m = pattern.matcher(in);
        if (m.matches()) {
            return m.group(m.groupCount());
        } else {
            return null;
        }
    }
    public static String getNextId(Iterable<TreeNode> workList, TreeNode parent,
			String type, String defaultPrefix) {
    	String newId = "00";
        String fullNewId = "00";
        boolean found = false;

        for (int i = 0; i < parent.sizeChildren(); i++) {
            if (parent.getChild(i).getType().equals(type) || type == null) {
                String number = getNumberStringFromId(parent.getChild(i).getId());
                if (number == null) {
                    continue;
                }
                try {
                    if (Integer.parseInt(number) >= Integer.parseInt(newId)) {
                        fullNewId = parent.getChild(i).getId();
                        newId = number;
                        found = true;
                    }
                } catch (NumberFormatException e) {
                }
            }
        }

        if(workList!=null)
            for (TreeNode nd : workList) {
                if (nd.getType().equals(type) || type == null) {
                    String number = getNumberStringFromId(nd.getId());
                    if (number == null) {
                        continue;
                    }
                    try {
                        if (Integer.parseInt(number) >= Integer.parseInt(newId)) {
                            fullNewId = nd.getId();
                            newId = number;
                            found = true;
                        }
                    } catch (NumberFormatException e) {
                    }
                }
            }

        if (!found && defaultPrefix != null) {
            fullNewId = defaultPrefix + fullNewId;
        }

        do {
            fullNewId = incLastNumberInString(fullNewId);
        } while (parent.findChild(fullNewId) != null);
        return fullNewId;
	}
    public static String getNextId(TreeNode parent, String type, String defaultPrefix) {
        return getNextId(null, parent, type, defaultPrefix);
    }

    // Utilites
    public static String incLastNumberInString(String str) {
        String regex = "\\d*";
        String tmp = str;

        while (!tmp.matches(regex))
            tmp = tmp.substring(1);
        String fixture = str.substring(0, str.length() - tmp.length());
        String tail = String.valueOf(Integer.parseInt(tmp) + 1);

        if (tail.length() < tmp.length()) {
            tail = makeZeroString(tmp.length() - tail.length()) + tail;
        }
        return fixture + tail;
    }

    private static String makeZeroString(int zeroCount) {
        StringBuilder sb = new StringBuilder(zeroCount);
        for (int i = 0; i < zeroCount; i++)
            sb.append("0");
        return sb.toString();
    }

    private static boolean isWindowsLocalPath(String path) {
        if (path.length() >= 2
                && (path.charAt(0) >= 'a' && path.charAt(0) <= 'z' || path
                        .charAt(0) >= 'A' && path.charAt(0) <= 'Z')
                && (path.charAt(1) == ':' || path.charAt(1) == '|'))
            return true;
        return false;
    }

    private final static Pattern patternImages = Pattern.compile(  "<(img|object)\\s+[^>]*(src|data)\\s*=\\s*(\\\"|')([^\\\"']+)(?:\\3)[^>]*>");
    
    public static Pattern getImagePattern(){
    	return patternImages;
    }
    

    private static void cloneDir(File source, File dest){
    	if(!source.exists())
    		return;
    	for(File ch: source.listFiles()){
    		File newch = new File(dest.getAbsolutePath()+File.separator+ch.getName());
    		if (ch.isDirectory())
    			cloneDir(ch, newch);
			else
				try {
					RequalityCLI.copyfile(ch.getAbsolutePath(), newch.getAbsolutePath());
				} catch (IOException e) {
					e.printStackTrace();
				}
    	}
    }

    public static String copyImages(String descr, TreeNode n) {
        Matcher matcher = patternImages.matcher(descr);
        while (matcher.find()) {
            if(matcher.group(1).equals("img") && matcher.group(2).equals("src")
                    ||matcher.group(1).equals("object") && matcher.group(2).equals("data")){
                String src = matcher.group(4);
                String pathToImage = src;
                pathToImage = pathToImage.replaceAll("\\\\", "/");
                // DELETE "file:/" PREFIXES
                //  remote paths cases
                if (pathToImage.startsWith("file://")) {
                    // Windows' shares
                    if (pathToImage.startsWith("file://///")) {
                        pathToImage = pathToImage.substring(8);
                    } else
                        //  "file://C:/Folder/..."
                        if (isWindowsLocalPath(pathToImage.substring(7)) ||
                                // "file:///C:/Folder/...", "file:///home/..." - remove
                                // "file://"
                                pathToImage.startsWith("file:///")) {
                            pathToImage = pathToImage.substring(7);
                        }
                    //  else "file://compname/share", prefix will be deleted below
                }
                // local paths cases - "file:/C:/Folder/...", "file:/home/..." -
                //  remove "file:"
                if (pathToImage.startsWith("file:")) {
                    pathToImage = pathToImage.substring(5);
                }
                // delete excess "/" on Windows
                if (pathToImage.startsWith("/")
                    && isWindowsLocalPath(pathToImage.substring(1))) {
                    pathToImage = pathToImage.substring(1);
                }
                try {
                    // search source file
                	if((pathToImage.startsWith("http") || pathToImage.startsWith("ftp")) && pathToImage.contains("?"))
                		pathToImage = pathToImage.substring(0,pathToImage.indexOf("?"));
                    File srcFile = new File(pathToImage);
                    if (srcFile.isAbsolute()) {
                        if (!srcFile.exists()) { // in URI
                            String absPath = srcFile.getAbsolutePath();
                            absPath = absPath.replaceAll("\\\\", "/");
                            if (isWindowsLocalPath(absPath)) {
                                absPath = "/" + absPath;
                            }
                            if (absPath.startsWith("//")) { // Windows' shares
                                absPath = "///" + absPath;
                            }
                            srcFile = new File(URI.create("file:" + absPath));
                        }
                    } else {
                        //  suppose, file in URI format
                        String rootPath = n.getTreeDB().getRootNode().getResourcesPath();
                        if (rootPath.endsWith("_resources")) { // hack
                            rootPath = rootPath.substring(0, rootPath.length() - 10) + "root";
                        }
                        try {
                        	File f2 = new File(URI.create(rootPath + '/' + pathToImage));
                        	if(f2.exists())
                        		srcFile = new File(URI.create(rootPath + '/' + pathToImage));
                        } catch (IllegalArgumentException e) {
                            //  no, file was not URI
                            File rootFile = new File(URI.create(rootPath));
                            // pathToImage = pathToImage.replaceAll("/", File.separator);
                            srcFile = new File(rootFile + File.separator + pathToImage);
                        }
                    }
                    if (!srcFile.exists()) {
                     //   continue;
                    }
                    //  now srcFile is absolute not URI path
                    // short file name in URI (with %20 as ' ' etc.)
                    String shortDestFileName = (new File(srcFile.toURI().toURL().toString())).getName();
                    int nSuffix = 0;
                    //  divide FileName into File Name and extension
                    String ext = "";
                    int dot = shortDestFileName.lastIndexOf('.');
                    if (dot > -1) {
                        ext = shortDestFileName.substring(dot);
                        shortDestFileName = shortDestFileName.substring(0, dot);
                    }
                    //  check destination file
                    File dstFile = new File(URI.create(n.getResourcesPath() + '/' + shortDestFileName + ext));
                    if (srcFile.getAbsolutePath().equals(dstFile.getAbsolutePath())) {
                        continue;
                    }
                    while (dstFile.exists()) {
                        nSuffix++;
                        dstFile = new File(URI.create(n.getResourcesPath() + '/' + shortDestFileName + nSuffix + ext));
                    }
                    boolean skip = false;
                    //  copy
                    try {
                    	if(srcFile.exists()){
                    		FileInputStream in = new FileInputStream(srcFile);
                    		n.setResourceContent(dstFile.getName(), in);
                    		in.close();
                    	}else{
                    		try{
                    			File f = copyfile(n, src, dstFile.getAbsolutePath());
                    			dstFile = f;
                    		}
                    		catch(IOException ie){
                    			skip = true;
                    			//TODO copying fails
                    		}
                    	}
                    	if(!skip)
                    		descr = descr.replace(src, dstFile.toURI().toURL().toString().replace("\\", "/"));
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                } catch (MalformedURLException e) {
                    e.printStackTrace();
                }
            }
        }
        return descr;
    }
    public static String resolveRelativeImagesPath(String descr, TreeNode n) {
        if (!(n instanceof Location)) {
            return descr;
        }
        n = n.getParent();
        if (!(n instanceof Document)) {
            return descr;
        }
        Matcher matcher = patternImages.matcher(descr);
        while (matcher.find()) {
            if(matcher.group(1).equals("img")&&matcher.group(2).equals("src")
                    ||matcher.group(1).equals("object")&&matcher.group(2).equals("data")){
                String src = matcher.group(4);
                String pathToImage = src;
                pathToImage = pathToImage.replaceAll("\\\\", "/");
                pathToImage = n.getResourcesPath() + "/" + pathToImage;
                try{
                    if(new File(pathToImage).exists())
                        descr = descr.replace(src, pathToImage);
                }catch(Exception e){
                    //for tab images
                }
            }
        }
        return descr;
    }
    public static void deleteOldImages(Attribute attr, String descr, TreeNode n) {
    	if (attr == null || attr.getType() != AttributeType.STRING) {
            return;
        }
        deleteOldImages((String) attr.getValue(), descr, n,"", attr.getKey());
    }
    public static void deleteOldImages(String oldDescr, String descr, TreeNode n,String imgpref) {
    	deleteOldImages(oldDescr, descr, n, imgpref, null);
    }
    public static void deleteOldImages(String oldDescr, String descr, TreeNode n,String imgpref,String ignorekey) {
         //get old images list
    	Matcher matcher = patternImages.matcher(oldDescr);
        List<String> oldImages = new ArrayList<String>();
        while (matcher.find()) {
            if(matcher.group(1).equals("img")&&matcher.group(2).equals("src")
                    ||matcher.group(1).equals("object")&&matcher.group(2).equals("data")){
                String src = matcher.group(4);
                if(src.startsWith("data:image"))
                	continue;
                String s = imgpref+src;
                if(n != null)
                	s = src.replace("${node.resURL}", n.getResourcesPath());
                try{
                	if(s.startsWith("file:"))
                		s=s.substring(5);
                	if(isWindowsLocalPath(s.substring(1)))
                		s=s.substring(1);
                	boolean sidefile = new File(s).exists();
                	//TODO image handling in console mode
                    if(Activator.getDefault()!=null && sidefile){
                    	if(sidefile)
                    		src = s;
                    	sidefile= new File(src).exists();
                        oldImages.add(src);
                    }
                }catch(Exception e){
                	e.printStackTrace();
                    //for tab images
                }
            }
        }
        //get new images list
    	List<String> newImages = new ArrayList<String>();
        if(descr != null){
        	matcher = patternImages.matcher(descr);
        	while (matcher.find()) {
        		if(matcher.group(1).equals("img")&&matcher.group(2).equals("src")
        				||matcher.group(1).equals("object")&&matcher.group(2).equals("data")){
        			String ss = matcher.group(4);
        			if(ss.startsWith("file:"))
						ss=ss.substring(5);
					if(isWindowsLocalPath(ss.substring(1)))
						ss=ss.substring(1);
        			newImages.add(ss);
        		}
        	}
        }
        String specPath = "";
        //try to find a picture in string variables
        if(n != null){
        	for(String key:n.getAttributeKeys()){
        		Object o = n.getAttributeValue(key);
        		if(o == null || !(o instanceof String))
        			continue;
        		String s = (String)o;
        		if(key.equals(ignorekey)) continue;
        		matcher = patternImages.matcher(s);
        		while (matcher.find()) {
        			if(matcher.group(1).equals("img")&&matcher.group(2).equals("src")
        					||matcher.group(1).equals("object")&&matcher.group(2).equals("data"))
        				try {
        					String ss = new URL(n.processResourcesPath(matcher.group(4))).toURI().toString();
        					if(ss.startsWith("file:"))
        						ss=ss.substring(5);
        					if(isWindowsLocalPath(ss.substring(1)))
        						ss=ss.substring(1);
        					newImages.add(ss);
        				} catch (MalformedURLException e) {
						// 	TODO Auto-generated catch block
        					e.printStackTrace();
        				} catch (URISyntaxException e) {
						// 	TODO Auto-generated catch block
        					e.printStackTrace();
        				}
        		}
        	}
        	specPath = n.getResourcesPath();
        	if(specPath!=null){
        		if(specPath.startsWith("file:"))
        			specPath = specPath.substring(5);
            	if(isWindowsLocalPath(specPath.substring(1)))
            		specPath = specPath.substring(1);
        	}
        }
        //remove duplicate items
        for (String newImg : newImages) {
            oldImages.remove(newImg);
        }
        //rest items - elements to delete
        for (String oldImg : oldImages) {
           // oldImg = oldImg.replaceAll("\\\\", "/");  
           if (new File(oldImg).exists()) {
        	   if(n != null){
        		   
        		   boolean inpath = oldImg.contains(specPath);
        	   	   if (oldImg.contains("/") && inpath && 
               		new File(oldImg).exists()) {
                		oldImg = oldImg.substring(oldImg.lastIndexOf('/') + 1);
                		n.deleteResource(oldImg);
        	   	   }
        	   }else{
        		   //here we suppose not to have an undo
        		   //on such operation
        		   
        		   File f = new File(oldImg);
        		   f.delete();
        		   if(f.getParentFile().list().length == 0)
        			   f.getParentFile().delete();
        	   } 
            }
        }
    }
    public static String getReportDescription(TreeNode n, String htmlFolderPath,
            String description, String prefix, String postfix, String basePath) {
    	if(description == null)
    		return description;
       	Matcher matcher = patternImages.matcher(description);
        while (matcher.find()) {
    		if(matcher.group(1).equals("img")&&matcher.group(2).equals("src")
    				||matcher.group(1).equals("object")&&matcher.group(2).equals("data"))
    		{
    			String mg0 = matcher.group(0),src = matcher.group(4),newsrc = src;
    			String path = n.processResourcesPath(src);
    	        if(path.startsWith("data:image"))
    	           	continue;
    	        if(path.startsWith("file:"))
    	        	path = path.substring(5);
    	        if(isWindowsLocalPath(path.substring(1)))
    	           	path = path.substring(1);
    	        path = path.replace("/", File.separator);
    	        path = path.replace("\\", File.separator);
    	        try {
					path = URLDecoder.decode(path, "UTF-8");
				} catch (UnsupportedEncodingException e) {
					e.printStackTrace();
				}
    	        File f = new File(path), alt = new File(htmlFolderPath + File.separator + path);
    	        if (!f.exists() && alt.exists())
    	           	f = alt;
    			String shortPath = "images"+ File.separator +  n.getUUId() + File.separator + path.substring(path.lastIndexOf(File.separator)+1);
    	        if(!f.exists()){
    	        	try{
    	        		File dest = RequalityCLI.copyfile(null, path, basePath + File.separator + shortPath);
    	        		if(dest.exists())
    	        			newsrc = shortPath.replace("\\","/");
    	        	}catch(IOException ex){
    	        		ex.printStackTrace();
    	        	}
    	        }else {
    	        	try{
    	        		String newPath = basePath + File.separator + shortPath;
    	        		File dest = RequalityCLI.copyfile(null, f.getAbsolutePath(), newPath);
    	        		if(dest.exists())
    	        			newsrc = shortPath.replace("\\","/");
    	        	}catch(IOException ex){
    	        		ex.printStackTrace();
    	        	}
    	        }
    	        if(!src.equals(newsrc))
    	        	description = description.replace(mg0,mg0.replace(src, newsrc));
    		}
        }
        return description;        
    }

	public static void setFileContent(File file, String content) {
        try {
            FileWriter fw = new FileWriter(file);
            fw.write(content);
            fw.close();
        } catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static void duplicateImages(TreeNode node, String path) {
    	String spath = node.getResourcesPath();
    	if(spath == null)
    		return;
    	if(spath.startsWith("file:"))spath = spath.substring(5);
    	if(path.startsWith("file:"))path = path.substring(5);
    	File f = new File(spath);
    	if(f.exists())
    		for(File f2: f.listFiles())
    			if(f2.isDirectory()){
    				File dest = new File(path+File.separator+f2.getName());
    				cloneDir(f2, dest);
    			}
	}

}
